/**
 * Bull SAS / OW2 Orchestra
 * Copyright (C) 2007 Bull S.A.S, and individual contributors as indicated
 * by the @authors tag.
 *
 * Bull, Rue Jean Jaures, B.P.68, 78340, Les Clayes-sous-Bois
 * This library is free software; you can redistribute it and/or modify it under the terms
 * of the GNU Lesser General Public License as published by the Free Software Foundation
 * version 2.1 of the License.
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 * You should have received a copy of the GNU Lesser General Public License along with this
 * program; if not, write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth
 * Floor, Boston, MA  02110-1301, USA.
 **/
/*
 * JBoss, Home of Professional Open Source
 * Copyright 2005, JBoss Inc., and individual contributors as indicated
 * by the @authors tag. See the copyright.txt in the distribution for a
 * full listing of individual contributors.
 *
 * This is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as
 * published by the Free Software Foundation; either version 2.1 of
 * the License, or (at your option) any later version.
 *
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this software; if not, write to the Free
 * Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301 USA, or see the FSF site: http://www.fsf.org.
 */
package org.ow2.orchestra.definition;

import java.net.URL;
import java.util.HashMap;
import java.util.Map;
import java.util.UUID;

import javax.xml.namespace.QName;

import org.jbpm.pvm.env.Environment;
import org.jbpm.pvm.internal.model.ExecutionImpl;
import org.jbpm.pvm.internal.model.ProcessDefinitionImpl;
import org.jbpm.pvm.internal.xml.Parse;
import org.ow2.orchestra.exception.StaticAnalysisException;
import org.ow2.orchestra.facade.data.def.ProcessDefinitionData;
import org.ow2.orchestra.facade.uuid.ProcessDefinitionUUID;
import org.ow2.orchestra.facade.uuid.ProcessInstanceUUID;
import org.ow2.orchestra.runtime.BpelExecution;
import org.ow2.orchestra.services.MutexRepository;
import org.ow2.orchestra.services.OperationKey;
import org.ow2.orchestra.services.ReceivingElement;
import org.ow2.orchestra.services.itf.Repository;
import org.ow2.orchestra.util.StaticAnalysisFault;
import org.ow2.orchestra.util.StyleSheetRepository;
import org.ow2.orchestra.wsdl.WsdlsInfos;

/**
 * a bpel process definition.
 *
 * Both the initial node in the process and the activity of the aggregated scope
 * should point to the same node which is the top level activity in the process.
 *
 * @author Goulven Le Jeune
 *
 */
public class BpelProcess extends ProcessDefinitionImpl {

  /** serialVersionUID. */
  private static final long serialVersionUID = 1L;

  private URL bpelURL;

  /**
   * specifies the query language used in the process for selection of nodes in
   * assignment. The default value for this attribute is:
   * "urn:oasis:names:tc:wsbpel:2.0:sublang:xpath1.0", which represents the
   * usage of [XPath 1.0] within WS-BPEL 2.0
   */
  protected String queryLanguage;

  /**
   * specifies the expression language used in the process element. The default
   * value for this attribute is:
   * "urn:oasis:names:tc:wsbpel:2.0:sublang:xpath1.0", which represents the
   * usage of [XPath 1.0] within WS-BPEL 2.0.
   */
  protected String expressionLanguage;

  /**
   * suppressJoinFailure attribute of the process element. determines whether
   * the joinFailure fault will be suppressed for all activities in the process.
   * The effect of the attribute at the process level can be overridden by an
   * activity using a different value for the attribute. The default for this
   * attribute is "no" at the process level. When this attribute is not
   * specified for an activity, it inherits its value from its closest enclosing
   * activity or from the process element if no enclosing activity specifies
   * this attribute.
   */
  protected boolean suppressJoinFailure = false;

  /** exitOnStandardFault attribute of the process element. */
  protected boolean exitOnStandardFault = false;

  /** targetNamespace attribute of the process element. */
  protected String targetNamespace = null;

  protected long instanceCount;

  // TODO replace with process deploy number.
  protected String processUUID;

  /**
   * Infos collected in wsdl files used by the process like portTypes,
   * messages...
   */
  protected WsdlsInfos wsdlInfos;

  /** styleSheetRepository. */
  protected StyleSheetRepository styleSheetRepository;

  protected Map<OperationKey, ReceivingElement> startElements;

  public BpelProcess(final URL bpelURL, final String processUUID) {
    this.bpelURL = bpelURL;
    instanceCount = 0;
    this.processUUID = processUUID;
  }

  /** For persistence * */
  protected BpelProcess() {
    super();
  }

  @Override
  protected ExecutionImpl newProcessInstance() {
    instanceCount++;
    ProcessInstanceUUID processInstanceUUID = new ProcessInstanceUUID(getUUID(), instanceCount);
    BpelExecution instance = new BpelExecution(processInstanceUUID, UUID.randomUUID().toString(), instanceCount);
    instance.setProcessDefinition(this);
    instance.setProcessInstance(instance);
    Environment.getCurrent().get(Repository.class).storeInstance(instance);
    MutexRepository.createInstanceLock(instance.getUUID());
    return instance;
  }


  public void addStartElement(final OperationKey operationKey, final ReceivingElement receivingElement,
      final boolean hasCorrelations) {
    if (startElements == null) {
      startElements = new HashMap<OperationKey, ReceivingElement>();
    }
    startElements.put(operationKey, receivingElement);
  }

  public ReceivingElement getStartElement(final OperationKey operationKey) {
    if (startElements != null) {
      return startElements.get(operationKey);
    }
    return null;
  }

  public String getExpressionLanguage() {
    return expressionLanguage;
  }

  public void setExpressionLanguage(final String expressionLanguage) {
    this.expressionLanguage = expressionLanguage;
  }

  public String getQueryLanguage() {
    return queryLanguage;
  }

  public void setQueryLanguage(final String queryLanguage) {
    this.queryLanguage = queryLanguage;
  }

  public boolean isSuppressJoinFailure() {
    return suppressJoinFailure;
  }

  public void setSuppressJoinFailure(final boolean suppressJoinFailure) {
    this.suppressJoinFailure = suppressJoinFailure;
  }

  public String getTargetNamespace() {
    return this.targetNamespace;
  }

  public void setTargetNamespace(final String targetNamespace) {
    this.targetNamespace = targetNamespace;
  }

  public void setWsdlInfos(final WsdlsInfos wsdlInfos) {
    this.wsdlInfos = wsdlInfos;
  }

  public WsdlsInfos getWsdlInfos() {
    return wsdlInfos;
  }

  public QName getQName() {
    return new QName(targetNamespace, name);
  }

  public StyleSheetRepository getStyleSheetRepository() {
    return styleSheetRepository;
  }

  public void setStyleSheetRepository(final StyleSheetRepository styleSheetRepository) {
    this.styleSheetRepository = styleSheetRepository;
  }

  public URL getBpelURL() {
    return bpelURL;
  }

  public void staticAnalysisCheck(final Parse parse) {
    if (startElements == null) {
      // SA00015
      String message = StaticAnalysisFault.getMessage("SA00015",
          "  Process should have at least one receive or pick with createInstance set to yes");
      parse.addProblem(message, new StaticAnalysisException("SA00015"));
    }
  }

  public ProcessDefinitionUUID getUUID() {
    return new ProcessDefinitionUUID(getQName(), processUUID);
  }

  public ProcessDefinitionData getProcessDefinitionData() {
    return new ProcessDefinitionData(getUUID(), name, targetNamespace);
  }

}
