/**
 * Copyright (C) 2007  Bull S. A. S.
 * Bull, Rue Jean Jaures, B.P.68, 78340, Les Clayes-sous-Bois
 * This library is free software; you can redistribute it and/or modify it under the terms
 * of the GNU Lesser General Public License as published by the Free Software Foundation
 * version 2.1 of the License.
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 * You should have received a copy of the GNU Lesser General Public License along with this
 * program; if not, write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth
 * Floor, Boston, MA  02110-1301, USA.
 **/
package org.ow2.orchestra.definition.activity;

import java.util.ArrayList;
import java.util.Date;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.logging.Level;
import java.util.logging.Logger;

import org.jbpm.pvm.Execution;
import org.jbpm.pvm.activity.ActivityExecution;
import org.jbpm.pvm.activity.ExternalActivity;
import org.ow2.orchestra.definition.BpelProcess;
import org.ow2.orchestra.definition.element.Catch;
import org.ow2.orchestra.definition.element.Link;
import org.ow2.orchestra.definition.element.Source;
import org.ow2.orchestra.definition.element.Target;
import org.ow2.orchestra.exception.BpelFaultException;
import org.ow2.orchestra.exception.faults.JoinFailure;
import org.ow2.orchestra.facade.data.def.ActivityDefinitionData;
import org.ow2.orchestra.facade.data.runtime.ActivityInstanceData;
import org.ow2.orchestra.facade.exception.OrchestraRuntimeException;
import org.ow2.orchestra.facade.runtime.ActivityType;
import org.ow2.orchestra.facade.uuid.ActivityDefinitionUUID;
import org.ow2.orchestra.facade.uuid.ActivityInstanceUUID;
import org.ow2.orchestra.lang.JoinCondition;
import org.ow2.orchestra.runtime.BpelExecution;
import org.ow2.orchestra.runtime.FlowRuntime;
import org.ow2.orchestra.runtime.LinkRuntime;
import org.ow2.orchestra.runtime.ScopeRuntime;
import org.ow2.orchestra.runtime.VariableRuntime;
import org.ow2.orchestra.services.MutexRepository;
import org.ow2.orchestra.services.MutexRepository.InstanceLock;
import org.ow2.orchestra.util.EnvTool;
import org.ow2.orchestra.var.Message;
import org.w3c.dom.Element;

/**
 * common parts of all the bpel activities.
 *
 * @author Tom Baeyens, Charles Souillard
 */
public abstract class AbstractActivity implements ExternalActivity {

  /** log. */
  private static Logger log = Logger.getLogger(AbstractActivity.class.getName());

  /** For persistence */
  protected long dbid;

  /** enclosing scope. */
  protected Scope enclosingScope;

  /** name attribute present for all activities. */
  protected String name = null;

  /** suppressJoinFailure attribute present for all activities. */
  protected boolean suppressJoinFailure;

  /**
   * targetsJoinCondition attribute set if this activity has target(s) and the
   * joinCondition attribute is set on targets element.
   */
  protected JoinCondition targetsJoinCondition = null;

  /** list of all sources present for all activities. */
  protected List<Source> sources = null;

  /** list of all targets present for all activities. */
  protected List<Target> targets = null;

  protected ActivityDefinitionUUID uuid;

  public long getDbid() {
    return this.dbid;
  }

  protected AbstractActivity() {
    super();
  }

  public enum TargetBehaviour {
    /**
     * all targets are evaluated and transitionCondition is true
     */
    execute,
    /**
     * All targets are not yet evaluated
     */
    unresolved,
    /**
     * all targets are evaluated and transitionCondition is false
     */
    notexecute
  }

  public abstract ActivityType getType();

  /**
   * Hook start of activity.
   *
   * @param bpelExecution
   */
  public void beforeRunning(final BpelExecution bpelExecution) {

    long activityCount = bpelExecution.getNextActivityCount();
    ActivityInstanceUUID activityInstanceUUID = new ActivityInstanceUUID(bpelExecution.getProcessInstanceUUID(),
      activityCount, name);
    bpelExecution.setStartedDate(new Date());
    bpelExecution.setActivityInstanceUUID(activityInstanceUUID);
    bpelExecution.setActivityPosition(activityCount);

    EnvTool.getRecorder().recordActivityStarted(createRuntimeData(bpelExecution));

  }

  /**
   * FIXME: this method should be abstract in future
   * @return a definition data
   */
  public ActivityDefinitionData createDefinitionData(final BpelProcess bpelProcess) {
    return new ActivityDefinitionData(bpelProcess.getProcessDefinitionData(), uuid, null,
      getType(), name);
  }

  /**
   * @param bpelExecution the execution
   * @return an instance data
   */
  public final ActivityInstanceData createRuntimeData(final BpelExecution bpelExecution) {
    return new ActivityInstanceData(createDefinitionData(bpelExecution.getProcessDefinition()), bpelExecution
      .getProcessInstance().getProcessInstanceData(), bpelExecution.getActivityInstanceUUID(), bpelExecution
      .getEncloserActivityInstanceUUID());
  }

  /**
   * Hook end of activity.
   *
   * @param bpelExecution
   */
  public void afterRunned(final BpelExecution bpelExecution) {

    EnvTool.getRecorder().recordActivityEnded(createRuntimeData(bpelExecution));
  }

  /**
   * Execute method called by process virtual machine.
   *
   * @param execution current execution
   */
  public void execute(final ActivityExecution execution) {

    final BpelExecution bpelExecution = (BpelExecution) execution;

    beforeRunning(bpelExecution);

    final TargetBehaviour targetBehaviour = targetsOK(bpelExecution);
    if (targetBehaviour == TargetBehaviour.execute) {
      if (log.isLoggable(Level.FINE)) {
        log.fine("### Executing " + execution.getNode().getName() + " in execution " + execution);
      }
      try {
        executeActivity(bpelExecution);
      } catch (final Exception e) {

        log.finest("ERROR CATCHED AbstractActivity: " + this);
        boolean find = false;

        // trace the corresponding fault Handler
        final Iterator<ScopeRuntime> it = bpelExecution.getScopeRuntimes().iterator();
        ScopeRuntime scopeRuntime = null;

        myWhile: while (it.hasNext()) {
          scopeRuntime = it.next();

          // get the fault handler of this scope
          final List<Catch> catchs = scopeRuntime.getScopeDefinition().getCatchs();

          if ((catchs != null) && (!catchs.contains(scopeRuntime.isCatchHasAlreadyBeenMatched()))) {
            Message faultMsg = null;
            for (final Catch catch1 : catchs) {
              // match the fault
              faultMsg = catch1.match(bpelExecution, e);
              if (faultMsg != null) {
                // place the execution on the good scope
                ScopeRuntime currentScope = bpelExecution.getScopeRuntime();
                while (scopeRuntime != currentScope) {
                  currentScope = bpelExecution.popScopeRuntime();
                }

                if (currentScope != bpelExecution.getScopeRuntime()) {
                  bpelExecution.pushScopeRuntime(currentScope);
                }
                log.finest("catch matched = " + catch1);
                if (log.isLoggable(Level.FINE)) {
                  log.fine("Execution of catch: " + this);
                }

                if (catch1.getFaultVariable() != null) {
                  final VariableRuntime variableRuntime = currentScope
                    .createVariable(bpelExecution, catch1.getFaultVariable());
                  if (variableRuntime.getVariableDefinition().getMessageType() != null) {
                    variableRuntime.setValue(faultMsg, bpelExecution.getActivityInstanceUUID());
                  } else {
                    if (faultMsg.getParts().values().size() >= 1) {
                      final Element elmt = (Element) faultMsg.getParts().values().toArray()[0];
                      variableRuntime.setValue(elmt, true, bpelExecution.getActivityInstanceUUID());
                    }
                  }
                }
                if (bpelExecution.getExecutions() != null) {
                  for (final Execution exec : new ArrayList<Execution>(bpelExecution.getExecutions())) {
                    ((BpelExecution) exec).terminate();
                  }
                }
                bpelExecution.createExecution().execute(catch1.getActivity());
                find = true;
                break myWhile;
              }
            }
          }
          if (catchs.contains(scopeRuntime.isCatchHasAlreadyBeenMatched())) {
            scopeRuntime.setCatchHasAlreadyBeenMatched(null);
          }

        }
        if (!find) {
          if (e instanceof BpelFaultException) {
            throw (BpelFaultException) e;
          }
          throw new OrchestraRuntimeException(e);
        }
      }
    } else if (targetBehaviour != TargetBehaviour.notexecute) {
      bpelExecution.waitForSignal();
    }
  }

  /**
   * Signal method.
   *
   * @param execution execution
   * @param signal signal
   */
  public void signal(final ActivityExecution execution, final String signal, final Map<String, Object> parameters) {
    if (signal == null || !signal.equals("target")) {
      signal((BpelExecution) execution, signal, parameters);
    }
  }

  /**
   * Signal method.
   *
   * @param execution execution
   * @param signal signal
   */
  public void signal(final BpelExecution execution, final String signal, final Map<String, Object> parameters) {
    throw new OrchestraRuntimeException("Signal not supported.");
  }

  /**
   * Returns true if all target are evaluated and if joinCondition is true.
   *
   * @param execution execution
   * @return true if all target are evaluated and if joinCondition is true.
   */
  private TargetBehaviour targetsOK(final BpelExecution execution) {
    // verify that all links have been evaluated
    if (this.targets == null || this.targets.isEmpty()) {
      return TargetBehaviour.execute;
    }
    boolean tjcLogicalOR = false;
    for (final Target target : this.targets) {
      final InstanceLock instanceLock = MutexRepository.lockInstance(execution);
      try {
        final String linkName = target.getLinkName();
        final FlowRuntime fr = execution.getFlowRuntimeWithLink(linkName);
        final Boolean linkStatus = fr.getLinkStatus(linkName);
        if (linkStatus == null) {
          // a link is not already evaluated
          if (log.isLoggable(Level.FINE)) {
            log.fine(this.getClass() + "." + getName() + " targetsOK, waiting for link : " + linkName);
          }
          // add this execution as a waiting execution
          fr.addWaitingExecution(linkName, execution);
          return TargetBehaviour.unresolved;
        } else {
          tjcLogicalOR = tjcLogicalOR || linkStatus;
        }
      } finally {
        MutexRepository.unlockInstance(instanceLock);
      }
    }
    boolean tjc = tjcLogicalOR;
    if (this.targetsJoinCondition != null) {
      try {
        tjc = this.targetsJoinCondition.getEvaluator().evaluateBoolean(execution);
      } catch (final Exception e) {
        // TODO throw joinFailure
        e.printStackTrace();
        setAllUnsetLinks(execution);
        throw new JoinFailure("bpel:joinFailure : " + this.targetsJoinCondition.getText()
          + ", exception while evaluating the targetJoinCondition");
      }
    }

    if (!tjc) {
      if (log.isLoggable(Level.FINE)) {
        log.fine(this.getClass() + ".targetsOK : " + false);
      }
      if (!this.suppressJoinFailure) {
        if (this.targetsJoinCondition != null) {
          throw new JoinFailure("bpel:joinFailure : " + this.targetsJoinCondition.getText()
            + ", targetJoinCondition is false and suppressJoinFailure is no");
        } else {
          throw new JoinFailure("bpel:joinFailure : targetJoinCondition is false and suppressJoinFailure is no");
        }
      } else {
        if (this.sources != null) {
          for (final Source source : this.sources) {
            final InstanceLock instanceLock = MutexRepository.lockInstance(execution);
            try {
              final String linkName = source.getLinkName();
              final FlowRuntime fr = execution.getFlowRuntimeWithLink(linkName);
              final LinkRuntime lr = fr.getLink(linkName);
              if (lr.getStatus() == null) {
                lr.setStatus(false);
              }
              final BpelExecution execToWakeUp = fr.getWaitingExecution(linkName);
              if (execToWakeUp != null) {
                execToWakeUp.execute(execToWakeUp.getNode());
              }
            } finally {
              MutexRepository.unlockInstance(instanceLock);
            }
          }
        }
        return TargetBehaviour.notexecute;
      }
      // TODO
      // set outgoingLinks (sources) status to false according to links semantic
      // (in same scope or not...)
      // see transitionCondition
      // notify waitingExecutions
      // if suppressjoinFailure : nothing (do not execute activity)
      // if not suppressJoinFailure : throw bpel:joinfFailure
    }
    if (log.isLoggable(Level.FINE)) {
      log.fine(this.getClass() + "." + getName() + " targetsOK : " + true);
    }
    return TargetBehaviour.execute;
  }

  /**
   * Executes sources of this activity : called when this activity is finished.
   *
   * @param execution execution
   */
  public void executeSources(final BpelExecution execution) {
    // Sources are executed once the enclosing activity is completed.
    // In the special case of an enclosing scope, this is not required to have
    // a successful completion. In that case (enclosing scope abnormally
    // completed)
    // sources must be executed only if there is a faultHandler associated to
    // the
    // enclosing scope and if this faultHandler has terminated without throwing
    // a fault.

    // if an error occurs during the TC evaluation
    // This error does not prevent enclosing activity to be set at "completed".
    // This error must be managed by the enclosing scope.
    // If the target activity is inside the source activity enclosing scope,
    // UNSET must be set on the concerned link status. (nothing to do)
    // -> done
    // If the target activity is outside the source activity enclosing scope,
    // false must be set on the concerned link status.
    // -> done
    // If source activity is a scope, the error must be propagated to its
    // father.
    // All remaining outgoing links (sources) having target activity in source
    // activity enclosing scope
    // must not have their transitionCondition evaluated (nothing to do)
    // -> done
    // All remaining outgoing links (sources) having target activity outside
    // source activity enclosing scope
    // must not have their status set to false
    // -> done

    afterRunned(execution);

    if (log.isLoggable(Level.FINE)) {
      log.fine("Executing sources of activity : " + this.getClass().getName() + "." + this.name);
    }
    if (this.sources == null) {
      return;
    }
    for (final Source source : this.sources) {
      final String linkName = source.getLinkName();
      boolean tc = true;
      if (source.getTransitionCondition() != null) {
        try {
          tc = source.getTransitionCondition().getEvaluator().evaluateBoolean(execution);
        } catch (final Exception e) {
          // TODO : throw error to enclosing scope : warning if the source
          // activity is a scope, propagate to parent
          setAllUnsetLinks(execution);
          e.printStackTrace();
        }
      }
      final InstanceLock instanceLock = MutexRepository.lockInstance(execution);
      try {
        execution.setLinkStatus(linkName, tc);
        if (log.isLoggable(Level.FINE)) {
          log.fine("executeSources, notifying executions of the link : " + linkName + "(" + tc + ")");
        }
        final FlowRuntime fr = execution.getFlowRuntimeWithLink(linkName);
        final BpelExecution execToWakeUp = fr.getWaitingExecution(linkName);
        if (execToWakeUp != null) {
          execToWakeUp.execute(execToWakeUp.getNode());
          // execToWakeUp.execute(execToWakeUp.getNode(), "target");
        }
      } finally {
        MutexRepository.unlockInstance(instanceLock);
      }
    }
  }

  /**
   * For every LinkRuntime having status UNSET, set it to false if target
   * activity is not defined in source activity enclosing scope.
   *
   * @param execution execution
   */
  private void setAllUnsetLinks(final BpelExecution execution) {
    if (this.sources != null) {
      for (final Source source : this.sources) {
        final String linkName = source.getLinkName();
        final FlowRuntime fr = execution.getFlowRuntimeWithLink(linkName);
        final LinkRuntime lr = fr.getLink(linkName);
        if (lr.getStatus() == null) {
          final Link link = source.getLink();
          final Scope targetActivityEnclosingScope = link.getTargetActivity().getEnclosingScope();
          final Scope sourceActivityEnclosingScope = link.getSourceActivity().getEnclosingScope();
          if (!(targetActivityEnclosingScope == sourceActivityEnclosingScope)) {
            lr.setStatus(false);
          }
        }
      }
    }
  }

  /**
   * This method must be implemented by every activity. This is the business
   * behaviour of the activity.
   *
   * @param execution execution
   */
  protected abstract void executeActivity(BpelExecution execution);

  public Scope getEnclosingScope() {
    return this.enclosingScope;
  }

  public void setEnclosingScope(final Scope enclosingScope) {
    this.enclosingScope = enclosingScope;
  }

  public String getName() {
    return this.name;
  }

  public void setName(final String name) {
    this.name = name;
  }

  public boolean isSuppressJoinFailure() {
    return this.suppressJoinFailure;
  }

  public void setSuppressJoinFailure(final boolean suppressJoinFailure) {
    this.suppressJoinFailure = suppressJoinFailure;
  }

  public List<Source> getSources() {
    return this.sources;
  }

  public void setSources(final List<Source> sources) {
    this.sources = sources;
  }

  public List<Target> getTargets() {
    return this.targets;
  }

  public void setTargets(final List<Target> targets) {
    this.targets = targets;
  }

  public JoinCondition getTargetsJoinCondition() {
    return this.targetsJoinCondition;
  }

  public void setTargetsJoinCondition(final JoinCondition targetsJoinCondition) {
    this.targetsJoinCondition = targetsJoinCondition;
  }

  public ActivityDefinitionUUID getUuid() {
    return this.uuid;
  }

  public void setUuid(final ActivityDefinitionUUID uuid) {
    this.uuid = uuid;
  }

  @Override
  public int hashCode() {
    return this.uuid.hashCode();
  }

  @Override
  public boolean equals(final Object obj) {
    if (obj == null) {
      return false;
    }
    if (!(obj instanceof AbstractActivity)) {
      return false;
    }
    final AbstractActivity other = (AbstractActivity) obj;
    return other.getUuid().equals(this.uuid);
  }
}
