/**
 * Copyright (C) 2007  Bull S. A. S.
 * Bull, Rue Jean Jaures, B.P.68, 78340, Les Clayes-sous-Bois
 * This library is free software; you can redistribute it and/or modify it under the terms
 * of the GNU Lesser General Public License as published by the Free Software Foundation
 * version 2.1 of the License.
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 * You should have received a copy of the GNU Lesser General Public License along with this
 * program; if not, write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth
 * Floor, Boston, MA  02110-1301, USA.
 **/
package org.ow2.orchestra.definition.activity;

import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import org.jbpm.pvm.Execution;
import org.jbpm.pvm.model.Node;
import org.jbpm.pvm.model.OpenExecution;
import org.ow2.novabpm.util.Misc;
import org.ow2.orchestra.exception.faults.CompletionConditionFailure;
import org.ow2.orchestra.exception.faults.InvalidBranchCondition;
import org.ow2.orchestra.facade.data.runtime.ForeachInitializedData;
import org.ow2.orchestra.facade.runtime.ActivityType;
import org.ow2.orchestra.lang.Expression;
import org.ow2.orchestra.lang.evaluator.InvalidExpressionValue;
import org.ow2.orchestra.runtime.BpelExecution;
import org.ow2.orchestra.runtime.ForEachRuntime;
import org.ow2.orchestra.util.EnvTool;

/**
 * the runtime behaviour of the forEach activity.
 *
 * @author Tom Baeyens, Charles Souillard
 */
public class ForEach extends AbstractActivity {

  /** serialVersionUID. */
  private static final long serialVersionUID = 1L;

  private String counterName;

  private boolean parallel;

  private Expression startCounterValue;

  private Expression finalCounterValue;

  private Expression completionCondition;

  private boolean successfulBranchesOnly = false;

  public ForEach() {
    super();
  }

  @Override
  public void signal(final BpelExecution execution, final String signal,
      final Map<String, Object> parameters) {
    if (signal == null) {
      // this is the child execution returning
      // end child execution
      execution.end(Execution.STATE_ENDED);

      BpelExecution foreachExecution = execution.getParent();
      // remove finished execution from parent
      foreachExecution.removeExecution(execution);

      ForEachRuntime forEachRuntime = foreachExecution.getForEachRuntime();
      // count one more branch as successful
      forEachRuntime.addTerminatedBranch(true);
      // check if completion condition is true
      if (!forEachRuntime.isCompleted()) {
        // completion condition is not reached
        if (!parallel) {
          // Sequential ForEach
          // TODO: check if remaining executions can set completion condition to true
          // throw completionConditionFailure otherwise
          if (!sequentialFor(foreachExecution, forEachRuntime)) {
            // if false, no child has been created
            // it means completion condition can never be true
            throw new CompletionConditionFailure();
          }
        } else {
          // Parallel ForEach
          // foreachExecution is active only when all children are finished
          if (foreachExecution.isActive()) {
            // if this code is executed, it means completion condition can never be true
            throw new CompletionConditionFailure();
          }
        }
      } else {
        // Completion condition is true
        // stop for each execution

        if (parallel) {
          // Stop children executions
          // Remove other waiting executions
          Collection<OpenExecution> children = foreachExecution.getExecutions();
          if (children != null) {
            for (OpenExecution child : new ArrayList<OpenExecution>(children)) {
              BpelExecution bpelChild = (BpelExecution) child;
              bpelChild.terminate();
            }
          }
        }
        // signal must be non-null to make the distinction with child
        // executions
        // in the first line of this method
        foreachExecution.signal("all-finished");
      }

    } else if (signal.equals("all-finished")) {
      execution.popForEachRuntime();
      executeSources(execution);
    }
  }

  /**
   * business logic.
   *
   * @param execution
   *          execution
   */
  @Override
  public void executeActivity(final BpelExecution execution) {
    // Evaluate start counter value
    int startCounterRuntimeValue;
    try {
      startCounterRuntimeValue = startCounterValue.getEvaluator()
          .evaluateUnsignedInt(execution);
    } catch (Exception e) {
      throw new InvalidExpressionValue("startCounterValue : "
          + startCounterValue, e);
    }
    // Evaluate final counter value
    int finalCounterRuntimeValue;
    try {
      finalCounterRuntimeValue = finalCounterValue.getEvaluator()
          .evaluateUnsignedInt(execution);
    } catch (Exception e) {
      throw new InvalidExpressionValue("finalCounterValue : "
          + finalCounterValue);
    }

    // For each scope should execute n = N + 1 times
    int n = Math.max(0, finalCounterRuntimeValue - startCounterRuntimeValue + 1);


    Integer completionConditionValue = null;
    // Evaluate completion condition B
    int completionConditionRuntime;
    if (completionCondition != null) {
      try {
        completionConditionRuntime = completionCondition.getEvaluator()
            .evaluateUnsignedInt(execution);
      } catch (Exception e) {
        throw new InvalidExpressionValue("completionCondition : "
            + completionCondition, e);
      }
      // Check B > N + 1
      // reject with bpel:invalidBranchCondition otherwise
      if (completionConditionRuntime > n) {
        throw new InvalidBranchCondition();
      }
      completionConditionValue = completionConditionRuntime;
    } else {
      // when no completion condition is set,
      // wait for all branches (N + 1)
      completionConditionRuntime = n;
    }

    // Create ForEach runtime
    ForEachRuntime forEachRuntime = new ForEachRuntime(startCounterRuntimeValue,
        finalCounterRuntimeValue, completionConditionRuntime, successfulBranchesOnly);
    EnvTool.getRecorder().recordForeachInitialized(
      new ForeachInitializedData(execution.getActivityInstanceUUID(), startCounterRuntimeValue, finalCounterRuntimeValue,
        completionConditionValue));

    execution.pushForEachRuntime(forEachRuntime);

    // set execution to wait state
    execution.waitForSignal();

    // Execute for each scopes if completion condition is false
    if (completionConditionRuntime != 0) {
      if (parallel) {
        // launch all executions in parallel
        parallelFor(execution, forEachRuntime);
      } else {
        // launch one execution
        if (!sequentialFor(execution, forEachRuntime)) {
          // if returns false, no execution has been created
          // this is a bug
          Misc.unreachableStatement();
        }
      }
    } else {
      // if completion condition is true (or n = 0)
      // signal execution (no scope execution to create)
      execution.signal("all-finished");
    }
  }

  /**
   *
   * @param execution parent execution
   * @return a child execution
   */
  private BpelExecution createChildExecution(final BpelExecution execution, final int counterValue) {
    // create ADDITIONNAL_SCOPE_VARIABLES. The enclosed scope will use variables defined in
    // this map to add variables in the scope.
    // The variables are previously defined in binding.
    Map<String, Object> additionnalVariables = new HashMap<String, Object>();
    additionnalVariables.put(counterName, counterValue);

    BpelExecution childExecution = (BpelExecution) execution.createExecution();
    childExecution.createVariable(BpelExecution.ADDITIONNAL_SCOPE_VARIABLES, additionnalVariables);
    return childExecution;
  }


  /**
   * @param execution the execution
   * @param forEachRuntime the foreachRuntime
   * @return true if a child execution has been created
   */
  private boolean sequentialFor(final BpelExecution execution,
      final ForEachRuntime forEachRuntime) {
    Node node = execution.getNode();
    int counterValue = forEachRuntime.getNextCurrentCounterValue();
    if (counterValue != ForEachRuntime.NO_MORE_COUNTER_VALUE) {
      Node scope = node.getNodes().get(0);
      // go to enclosed scope
      createChildExecution(execution, counterValue).execute(scope);
      return true;
    }
    return false;
  }

  private void parallelFor(final BpelExecution execution, final ForEachRuntime forEachRuntime) {
    Node node = execution.getNode();
    Node scope = node.getNodes().get(0);

    List<BpelExecution> childBpelExecutions = new ArrayList<BpelExecution>();

    // create each child executions.
    int counterValue = forEachRuntime.getNextCurrentCounterValue();
    while (counterValue != ForEachRuntime.NO_MORE_COUNTER_VALUE) {
      childBpelExecutions.add(createChildExecution(execution, counterValue));
      counterValue = forEachRuntime.getNextCurrentCounterValue();
    }

    // launch child executions.
    for (BpelExecution childBpelExecution : childBpelExecutions) {
      if (execution.getExecutions().contains(childBpelExecution)) {
        childBpelExecution.execute(scope);
      }
    }

  }

  public String getCounterName() {
    return counterName;
  }

  public void setCounterName(final String counterName) {
    this.counterName = counterName;
  }

  public boolean isParallel() {
    return parallel;
  }

  public void setParallel(final boolean parallel) {
    this.parallel = parallel;
  }

  public Expression getStartCounterValue() {
    return startCounterValue;
  }

  public void setStartCounterValue(final Expression startCounterValue) {
    this.startCounterValue = startCounterValue;
  }

  public Expression getFinalCounterValue() {
    return finalCounterValue;
  }

  public void setFinalCounterValue(final Expression finalCounterValue) {
    this.finalCounterValue = finalCounterValue;
  }

  public Expression getCompletionCondition() {
    return completionCondition;
  }

  public void setCompletionCondition(final Expression completionCondition) {
    this.completionCondition = completionCondition;
  }

  public boolean isSuccessfulBranchesOnly() {
    return successfulBranchesOnly;
  }

  public void setSuccessfulBranchesOnly(final boolean successfulBranchesOnly) {
    this.successfulBranchesOnly = successfulBranchesOnly;
  }

  @Override
  public ActivityType getType() {
    return ActivityType.FOR_EACH;
  }

}
