/**
 * Copyright (C) 2007  Bull S. A. S.
 * Bull, Rue Jean Jaures, B.P.68, 78340, Les Clayes-sous-Bois
 * This library is free software; you can redistribute it and/or modify it under the terms
 * of the GNU Lesser General Public License as published by the Free Software Foundation
 * version 2.1 of the License.
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 * You should have received a copy of the GNU Lesser General Public License along with this
 * program; if not, write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth
 * Floor, Boston, MA  02110-1301, USA.
 **/
package org.ow2.orchestra.definition.activity;

import java.util.List;

import javax.xml.namespace.QName;

import org.ow2.orchestra.definition.BpelProcess;
import org.ow2.orchestra.definition.element.Copy;
import org.ow2.orchestra.definition.element.Correlation;
import org.ow2.orchestra.definition.element.Variable;
import org.ow2.orchestra.facade.data.def.ReplyActivityDefinitionData;
import org.ow2.orchestra.facade.exception.OrchestraRuntimeException;
import org.ow2.orchestra.facade.runtime.ActivityType;
import org.ow2.orchestra.runtime.BpelExecution;
import org.ow2.orchestra.runtime.CorrelationSetRuntime;
import org.ow2.orchestra.runtime.ScopeRuntime;
import org.ow2.orchestra.runtime.VariableRuntime;
import org.ow2.orchestra.services.OperationKey;
import org.ow2.orchestra.services.itf.Replier;
import org.ow2.orchestra.util.XmlConstants;
import org.ow2.orchestra.var.Message;

/**
 * the runtime behaviour of the reply activity.
 *
 * @author Charles Souillard, Tom Baeyens
 * @version $Id: Reply.java 2055 2009-01-26 11:22:19Z porcherg $
 */
public class Reply extends AutomaticActivity {

  /** serialVersionUID. */
  private static final long serialVersionUID = 1L;

  /** partnerLink attribute of this Reply. */
  protected String partnerLinkName = null;

  /** operation attribute of this Reply. */
  protected String operationName = null;

  /** portType attribute of this Reply. */
  protected QName portTypeQName = null;

  /** variable attribute of this Reply. */
  protected String variable = null;

  /** faultName attribute of this Reply. */
  protected QName faultQName = null;

  /** messageExchange attribute of this Reply. */
  protected String messageExchange = XmlConstants.BPEL_DEFAULT_MESSAGE_EXCHANGE;

  /** list of correlations of this Reply. */
  protected List<Correlation> correlations = null;

  protected List<Copy> outCopies;

  protected Variable anonymousOutboundMessageVariable;

  public static final String OUTBOUND_ANONYMOUS_WSDL_MESSAGE_NAME = "ORCHESTRA_4_OUTBOUND_WSDL_MESSAGE";

  public Reply() {
    super();
  }

  /**
   * business logic.
   *
   * @param execution
   *            execution
   */
  @Override
  public void executeBusinessLogic(final BpelExecution execution) {
    // first find the replier
    OperationKey operationKey = new OperationKey(execution.getProcessDefinition()
        .getQName(), portTypeQName, operationName);
    ScopeRuntime scopeRuntime = execution.getScopeRuntime();
    Replier replier = execution.getReplier(operationKey, messageExchange);

    if (replier == null) {
      throw new OrchestraRuntimeException("couldn't find a replier for " + this
          + " on operation " + operationKey);
    }

    // get the reply message
    QName msgType = null;
    Message reply = null;
    if (variable != null) {
      reply = (Message) execution.getVariableValue(variable);
      msgType = execution.getVariableByName(variable).getVariableDefinition().getMessageType();
    } else if (outCopies != null && !outCopies.isEmpty()) {
      scopeRuntime = execution.getScopeRuntime();
      VariableRuntime anonymousOutboundMessageVariableRuntime = scopeRuntime
          .createVariable(execution, this.anonymousOutboundMessageVariable);
      // toparts
      for (Copy copy : this.outCopies) {
        copy.execute(execution);
      }
      reply = (Message) anonymousOutboundMessageVariableRuntime.getValue();
      msgType = anonymousOutboundMessageVariable.getMessageType();
    }
    // check correlations
    if (correlations != null) {
      //we verify correlation
      for (Correlation correlation : correlations) {
        // Check correlation on outgoing message
        CorrelationSetRuntime correlationSetRuntime =
          execution.getCorrelationSetByName(correlation.getCorrelationSet().getName());
        correlationSetRuntime.checkCSValues(
            execution,
            reply,
            msgType,
            correlation.getInitiate());
      }
    }
    // do the reply
    replier.reply(reply);
  }

  public String getOperation() {
    return this.operationName;
  }

  public void setOperation(final String operation) {
    this.operationName = operation;
  }

  public String getPartnerLinkName() {
    return this.partnerLinkName;
  }

  public void setPartnerLinkName(final String partnerLinkName) {
    this.partnerLinkName = partnerLinkName;
  }

  public QName getPortTypeQName() {
    return this.portTypeQName;
  }

  public void setPortTypeQName(final QName portType) {
    this.portTypeQName = portType;
  }

  public String getVariable() {
    return this.variable;
  }

  public void setVariable(final String variable) {
    this.variable = variable;
  }

  public List<Correlation> getCorrelations() {
    return this.correlations;
  }

  public void setCorrelations(final List<Correlation> correlations) {
    this.correlations = correlations;
  }

  public QName getFaultQName() {
    return this.faultQName;
  }

  public void setFaultQName(final QName faultName) {
    this.faultQName = faultName;
  }

  public String getMessageExchange() {
    return this.messageExchange;
  }

  public void setMessageExchange(final String messageExchange) {
    this.messageExchange = messageExchange;
  }

  public void setOutCopies(final List<Copy> outCopies) {
    this.outCopies = outCopies;
  }

  public void setAnonymousOutboundMessageVariable(
      final Variable anonymousOutboundMessageVariable) {
    this.anonymousOutboundMessageVariable = anonymousOutboundMessageVariable;
  }

  @Override
  public ActivityType getType() {
    return ActivityType.REPLY;
  }

  @Override
  public ReplyActivityDefinitionData createDefinitionData(final BpelProcess bpelProcess) {
    return new ReplyActivityDefinitionData(bpelProcess.getProcessDefinitionData(),
      uuid, null, name, null, null, null);
  }

}
