package org.ow2.orchestra.env;

import java.util.HashMap;
import java.util.Map;

import org.jbpm.pvm.env.EnvironmentFactory;
import org.ow2.orchestra.facade.exception.OrchestraRuntimeException;
import org.ow2.orchestra.persistence.db.DbHistory;
import org.ow2.orchestra.persistence.db.DbJournal;
import org.ow2.orchestra.persistence.log.LoggerArchiver;
import org.ow2.orchestra.persistence.log.LoggerRecorder;
import org.ow2.orchestra.services.Archiver;
import org.ow2.orchestra.services.Querier;
import org.ow2.orchestra.services.Recorder;
import org.ow2.orchestra.services.handlers.FinishedInstanceHandler;
import org.ow2.orchestra.services.handlers.UndeployedProcessHandler;
import org.ow2.orchestra.services.handlers.impl.NoOpFinishedInstanceHandler;
import org.ow2.orchestra.services.handlers.impl.NoOpUndeployedProcessHandler;
import org.ow2.orchestra.services.impl.DbRepository;
import org.ow2.orchestra.services.impl.PublisherImpl;
import org.ow2.orchestra.services.impl.UnsupportedInvoker;
import org.ow2.orchestra.services.itf.Invoker;
import org.ow2.orchestra.services.itf.Publisher;
import org.ow2.orchestra.services.itf.Repository;

/**
 * Generates a default environment configuration for running Orchestra.
 *
 * @author Guillaume Porcher, Charles Souillard
 *
 */
public class DefaultEnvXMLGenerator {

  public static final String CORE_HB_CONF_NAME = "hibernate-configuration:core";
  public static final String CORE_HB_SESSION_FACTORY_NAME = "hibernate-session-factory:core";
  public static final String CORE_HB_SESSION_NAME = "hibernate-session:core";
  public static final String CORE_RUNTIME_SESSION_NAME = "runtime-session:core";
  public static final String CORE_QUERIER_SESSION_NAME = "querier-session:core";

  public static final String HISTORY_HB_CONF_NAME = "hibernate-configuration:history";
  public static final String HISTORY_HB_SESSION_FACTORY_NAME = "hibernate-session-factory:history";
  public static final String HISTORY_HB_SESSION_NAME = "hibernate-session:history";
  public static final String HISTORY_QUERIER_SESSION_NAME = "querier-session:history";

  private final Map<String, String> applicationVariables;

  private final Map<String, String> blockVariables;

  private Class< ? extends Repository> repositoryType = null;

  private Class< ? extends Publisher> publisherType = null;

  private Class< ? extends Invoker> invokerType = null;

  private String recorderXml = "";
  private String archiverXml = "";

  private String fihXml = "";
  private String uphXml = "";

  public DefaultEnvXMLGenerator() {
    this.applicationVariables = new HashMap<String, String>();
    this.blockVariables = new HashMap<String, String>();
    this.setPublisherType(PublisherImpl.class);
    this.setInvokerType(UnsupportedInvoker.class);

    /*
     * Runtime (Engine + journal) Hibernate configuration
     */
    String hibernateCoreConfigXml = "";
    hibernateCoreConfigXml += "    <hibernate-configuration name='" + CORE_HB_CONF_NAME + "'>\n";
    hibernateCoreConfigXml += "      <properties resource='hibernate.properties' />\n";
    hibernateCoreConfigXml += "      <mappings resource='bpel.hibernate.mappings.xml' />\n";
    hibernateCoreConfigXml += "      <cache-configuration resource='bpel.cache.xml' usage='read-write' />\n";
    hibernateCoreConfigXml += "    </hibernate-configuration>";

    this.setApplicationEntry(CORE_HB_CONF_NAME, hibernateCoreConfigXml);
    this.setApplicationEntry(CORE_HB_SESSION_FACTORY_NAME,
        "<hibernate-session-factory name='" + CORE_HB_SESSION_FACTORY_NAME
        + "' configuration='" + CORE_HB_CONF_NAME + "'/>");

    this.setBlockEntry(CORE_HB_SESSION_NAME,
        "    <hibernate-session name='" + CORE_HB_SESSION_NAME
        + "' factory='" + CORE_HB_SESSION_FACTORY_NAME + "' init='eager'/>");

    this.setBlockEntry(CORE_RUNTIME_SESSION_NAME,
        "    <hibernate-bpel-persistence-service name='" + CORE_RUNTIME_SESSION_NAME
        + "' session='" + CORE_HB_SESSION_NAME + "'/>");
    this.setBlockEntry(CORE_QUERIER_SESSION_NAME,
        "    <querier-db-session name='" + CORE_QUERIER_SESSION_NAME
        + "' session='" + CORE_HB_SESSION_NAME + "'/>");
    /*
     * History Hibernate configuration
     */
    String hibernateHistoryConfigXml = "";
    hibernateHistoryConfigXml += "    <hibernate-configuration name='" + HISTORY_HB_CONF_NAME + "'>\n";
    hibernateHistoryConfigXml += "      <properties resource='hibernate-history.properties' />\n";
    hibernateHistoryConfigXml += "      <mappings resource='bpel.hibernate.mappings.xml' />\n";
    hibernateHistoryConfigXml += "      <cache-configuration resource='bpel.cache.xml' usage='read-write' />\n";
    hibernateHistoryConfigXml += "    </hibernate-configuration>";

    this.setApplicationEntry(HISTORY_HB_CONF_NAME, hibernateHistoryConfigXml);
    this.setApplicationEntry(HISTORY_HB_SESSION_FACTORY_NAME,
        "<hibernate-session-factory name='" + HISTORY_HB_SESSION_FACTORY_NAME
        + "' configuration='" + HISTORY_HB_CONF_NAME + "'/>");

    this.setBlockEntry(HISTORY_HB_SESSION_NAME,
        "    <hibernate-session name='" + HISTORY_HB_SESSION_NAME
        + "' factory='" + HISTORY_HB_SESSION_FACTORY_NAME + "' init='eager'/>");

    this.setBlockEntry(HISTORY_QUERIER_SESSION_NAME,
        "    <querier-db-session name='" + HISTORY_QUERIER_SESSION_NAME
        + "' session='" + HISTORY_HB_SESSION_NAME + "'/>");

    /*
     * Other services
     */
    String commandServiceConfigXml = "    <command-service>\n";
    commandServiceConfigXml += "      <retry-interceptor />\n";
    commandServiceConfigXml += "      <environment-interceptor />\n";
    commandServiceConfigXml += "      <standard-transaction-interceptor />\n";
    commandServiceConfigXml += "    </command-service>\n";
    this.setApplicationEntry("command-service", commandServiceConfigXml);
    this.setApplicationEntry("job-executor",  "    <job-executor threads='1' auto-start='false' />");

    this.setRepositoryType(DbRepository.class);

    this.setBlockEntry("transaction", "    <transaction />");

    this.setBlockEntry("timer_session", "    <timer-session />");
    this.setBlockEntry("message_session", "    <bpel-message-session />");
    this.setBlockEntry("job-session", "    <job-db-session session='" + CORE_HB_SESSION_NAME + "' />");


    this.setRecorderType(LoggerRecorder.class);

    this.setFinishedInstanceHandlerType(NoOpFinishedInstanceHandler.class);
    this.setUndeployedProcessHandlerType(NoOpUndeployedProcessHandler.class);

    this.setArchiverType(LoggerArchiver.class);

    /*
     * Journal creation (CORE_QUERIER_SESSION is given to constructor)
     */
    this.setBlockEntry("journal", "<journal name='journal' class='" + DbJournal.class.getName() + "'>"
        + "<arg><ref object='" + CORE_QUERIER_SESSION_NAME + "' /></arg></journal>");

    /*
     * History creation (HISTORY_QUERIER_SESSION is given to constructor)
     */
    this.setBlockEntry("history", "<history name='history' class='" + DbHistory.class.getName() + "'>"
        + "<arg><ref object='" + HISTORY_QUERIER_SESSION_NAME + "' /></arg></history>");
  }

  /**
   * Generates a xml for the given environment configuration.
   *
   * @return
   */
  public String createEnvironmentXml() {
    if (this.invokerType == null) {
      throw new OrchestraRuntimeException("Error while creating environment xml : invoker is null");
    }
    if (this.publisherType == null) {
      throw new OrchestraRuntimeException("Error while creating environment xml : publisher is null");
    }
    if (this.repositoryType == null) {
      throw new OrchestraRuntimeException("Error while creating environment xml : repository is null");
    }
    this.setBlockEntry(Recorder.DEFAULT_KEY,
        "    <chainer name='" + Recorder.DEFAULT_KEY + "'>"
        + this.recorderXml
        + "<ref object='journal' />"
        + "</chainer>");

    this.setBlockEntry(Archiver.DEFAULT_KEY,
        "    <chainer name='" + Archiver.DEFAULT_KEY + "'>"
        + this.archiverXml
        + "<ref object='history' />"
        + "</chainer>");

    this.setBlockEntry(Querier.DEFAULT_KEY,
        "    <queryApi name='" + Querier.DEFAULT_KEY + "'>"
        + "<ref object='journal' />"
        + "<ref object='history' />"
        + "</queryApi>");

    this.setBlockEntry(FinishedInstanceHandler.DEFAULT_KEY, this.fihXml);
    this.setBlockEntry(UndeployedProcessHandler.DEFAULT_KEY, this.uphXml);

    String result = "<environment-definition>\n\n"
      + "  <environment-factory>\n";
    for (final String entry : this.applicationVariables.values()) {
      result += entry + "\n";
    }
    result += "    <repository class='" + this.repositoryType.getName() + "' />\n"
      + "    <publisher class='" + this.publisherType.getName() + "' />\n"
      + "    <invoker class='" + this.invokerType.getName() + "' />\n"
      + "  </environment-factory>\n\n" + "  <environment>\n";
    for (final String entry : this.blockVariables.values()) {
      result += entry + "\n";
    }
    result += "  </environment>\n\n" + "</environment-definition>\n";
    return result;
  }

  /**
   * Creates a new EnvironmentFactory from the list of entries.
   */
  public EnvironmentFactory createEnvironmentFactory() {
    final EnvironmentFactory environmentFactory = BpelEnvironmentParser
    .parseEnvironmentFactoryFromXmlString(createEnvironmentXml());
    return environmentFactory;
  }

  /**
   * Return a string that declares an object to be added in the context.
   *
   * @param key
   *            objectName
   * @param classToUse
   *            classToUse at runtime
   * @return a string that declares an object to be added in the context.
   */
  @SuppressWarnings("unchecked")
  private static String getObjectDecl(final String key, final Class classToUse) {
    return "    <object name='" + key + "' class='" + classToUse.getName()
      + "' init='eager'/>";
  }

  /**
   * Return a string that declares an object to be added in the context.
   *
   * @param key
   *            objectName
   * @param classToUse
   *            classToUse at runtime
   * @param fieldKeys
   *            fields name
   * @param refKeys
   *            the keys of objects to reference in fields declaration
   * @return a string that declares an object to be added in the context.
   */
  @SuppressWarnings("unchecked")
  private static String getObjectDeclWithFields(final String key, final Class classToUse,
      final String[] fieldKeys, final String[] refKeys) {
    if (fieldKeys.length != refKeys.length) {
      throw new RuntimeException(
        "Error : fieldKeys and refKeys don't have the same number of elements");
    }
    String result = "    <object name='" + key + "' class='"
      + classToUse.getName() + "' init='eager'>\n";
    for (int i = 0; i < fieldKeys.length; i++) {
      result += "      <field name='" + fieldKeys[i] + "'><ref object='"
        + refKeys[i] + "' /></field>\n";
    }
    result += "    </object>";
    return result;
  }

  /**
   * Add or replace an entry in the environment to generate.
   *
   * @param name
   *            name of the entry
   * @param xml
   *            declaration of this entry
   * @return previous declaration for this name
   */
  public String setApplicationEntry(final String name, final String xml) {
    return this.applicationVariables.put(name, xml);
  }

  /**
   * Add or replace an entry in the environment to generate.
   *
   * @param name
   *            name of the entry
   * @param classToUse
   *            implementation
   * @return previous declaration for this name
   */
  public String setApplicationEntry(final String name, final Class< ? > classToUse) {
    final String xml = getObjectDecl(name, classToUse);
    return setApplicationEntry(name, xml);
  }

  /**
   * Add or replace an entry in the environment to generate.
   *
   * @param name
   *            name of the entry
   * @param classToUse
   *            implementation
   * @param fieldKeys
   *            list of fields to set
   * @param refKeys
   *            value to set to the fields
   * @return previous declaration for this name
   */
  public String setApplicationEntry(final String name, final Class< ? > classToUse,
      final String[] fieldKeys, final String[] refKeys) {
    final String xml = getObjectDeclWithFields(name, classToUse, fieldKeys, refKeys);
    return setApplicationEntry(name, xml);
  }

  /**
   * Add or replace an entry in the environment to generate.
   *
   * @param name
   *            name of the entry
   * @param xml
   *            declaration of this entry
   * @return previous declaration for this name
   */
  public String setBlockEntry(final String name, final String xml) {
    return this.blockVariables.put(name, xml);
  }

  /**
   * Add or replace an entry in the environment to generate.
   *
   * @param name
   *            name of the entry
   * @param classToUse
   *            implementation
   * @return previous declaration for this name
   */
  public String setBlockEntry(final String name, final Class< ? > classToUse) {
    final String xml = getObjectDecl(name, classToUse);
    return setBlockEntry(name, xml);
  }

  /**
   * Add or replace an entry in the environment to generate.
   *
   * @param name
   *            name of the entry
   * @param classToUse
   *            implementation
   * @param fieldKeys
   *            list of fields to set
   * @param refKeys
   *            value to set to the fields
   * @return previous declaration for this name
   */
  public String setBlockEntry(final String name, final Class< ? > classToUse,
      final String[] fieldKeys, final String[] refKeys) {
    final String xml = getObjectDeclWithFields(name, classToUse, fieldKeys, refKeys);
    return setBlockEntry(name, xml);
  }

  public void setRepositoryType(final Class< ? extends Repository> repositoryType) {
    this.repositoryType = repositoryType;
  }

  public void setPublisherType(final Class< ? extends Publisher> publisherType) {
    this.publisherType = publisherType;
  }

  public void setInvokerType(final Class< ? extends Invoker> invokerType) {
    this.invokerType = invokerType;
  }

  public void setRecorderType(final Class< ? extends Recorder>... recorderClasses) {
    this.recorderXml = "";
    for (final Class< ? extends Recorder> recorderClass : recorderClasses) {
      this.recorderXml += "<recorder class='" + recorderClass.getName() + "' />";
    }
  }

  public void setArchiverType(final Class< ? extends Archiver>... archiverClasses) {
    this.archiverXml = "";
    for (final Class< ? extends Archiver> archiverClass : archiverClasses) {
      this.archiverXml += "<archiver class='" + archiverClass.getName() + "' />";
    }
  }

  public void setFinishedInstanceHandlerType(final Class< ? extends FinishedInstanceHandler>... fihClasses) {
    this.fihXml = "<chainer name='" + FinishedInstanceHandler.DEFAULT_KEY + "'>";
    for (final Class< ? extends FinishedInstanceHandler> fihClass : fihClasses) {
      this.fihXml += "<finished-instance-handler class='" + fihClass.getName() + "' />";
    }
    this.fihXml += "</chainer>";
  }

  public void setUndeployedProcessHandlerType(final Class< ? extends UndeployedProcessHandler>... uphClasses) {
    this.uphXml = "<chainer name='" + UndeployedProcessHandler.DEFAULT_KEY + "'>";
    for (final Class< ? extends UndeployedProcessHandler> uphClass : uphClasses) {
      this.uphXml += "<undeployed-process-handler class='" + uphClass.getName() + "' />";
    }
    this.uphXml += "</chainer>";
  }

}
