/**
 * Copyright (C) 2006  Bull S. A. S.
 * Bull, Rue Jean Jaures, B.P.68, 78340, Les Clayes-sous-Bois
 * This library is free software; you can redistribute it and/or modify it under the terms
 * of the GNU Lesser General Public License as published by the Free Software Foundation
 * version 2.1 of the License.
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 * You should have received a copy of the GNU Lesser General Public License along with this
 * program; if not, write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth
 * Floor, Boston, MA  02110-1301, USA.
 **/
package org.ow2.orchestra.parsing.binding;

import java.util.ArrayList;
import java.util.List;
import java.util.logging.Level;
import java.util.logging.Logger;

import javax.xml.namespace.QName;

import org.jbpm.pvm.internal.model.NodeImpl;
import org.jbpm.pvm.internal.util.XmlUtil;
import org.jbpm.pvm.internal.xml.Parse;
import org.jbpm.pvm.internal.xml.Parser;
import org.ow2.orchestra.definition.BpelProcess;
import org.ow2.orchestra.definition.activity.Pick;
import org.ow2.orchestra.definition.element.Copy;
import org.ow2.orchestra.definition.element.OnAlarm;
import org.ow2.orchestra.definition.element.OnMessage;
import org.ow2.orchestra.definition.element.PartnerLink;
import org.ow2.orchestra.exception.StaticAnalysisException;
import org.ow2.orchestra.facade.def.ForDefinition;
import org.ow2.orchestra.facade.def.UntilDefinition;
import org.ow2.orchestra.facade.def.full.ActivityFullDefinition;
import org.ow2.orchestra.facade.def.full.OnAlarmFullDefinition;
import org.ow2.orchestra.facade.def.full.OnMessageFullDefinition;
import org.ow2.orchestra.facade.def.full.PickActivityFullDefinition;
import org.ow2.orchestra.facade.def.full.impl.OnAlarmFullDefinitionImpl;
import org.ow2.orchestra.facade.def.full.impl.OnMessageFullDefinitionImpl;
import org.ow2.orchestra.facade.def.full.impl.PickActivityFullDefinitionImpl;
import org.ow2.orchestra.facade.def.impl.ForDefinitionImpl;
import org.ow2.orchestra.facade.def.impl.UntilDefinitionImpl;
import org.ow2.orchestra.facade.runtime.ActivityType;
import org.ow2.orchestra.facade.uuid.ActivityDefinitionUUID;
import org.ow2.orchestra.facade.uuid.ProcessDefinitionUUID;
import org.ow2.orchestra.lang.Expression;
import org.ow2.orchestra.services.OperationKey;
import org.ow2.orchestra.util.BpelUtil;
import org.ow2.orchestra.util.StaticAnalysisFault;
import org.w3c.dom.Element;

public class PickBinding extends ActivityBinding {

  /** log. */
  private static Logger log = Logger.getLogger(PickBinding.class.getName());

  public PickBinding() {
    super("pick", ActivityType.PICK);
  }

  public Object parse(final Element element, final Parse parse, final Parser parser) {
    Pick pick = new Pick();
    setEnclosingScope(pick, parse);
    parseStandardAttributes(element, pick, parse);

    // Definition
    DefinitionKeeper definitionKeeper = parse.findObject(DefinitionKeeper.class);
    PickActivityFullDefinition pickActivityFullDefinition = (PickActivityFullDefinition) definitionKeeper
      .getActivityFullDefinition();
    List<OnMessageFullDefinition> onMessageFullDefinitionList = new ArrayList<OnMessageFullDefinition>();
    List<OnAlarmFullDefinition> onAlarmFullDefinitionList = new ArrayList<OnAlarmFullDefinition>();

    parseStandardElements(element, pick, parse);

    pick.setCreateInstance(BpelUtil.isCreateInstance(element));

    NodeImpl pickNode = createActivityNode(element, parse, parser, pick);

    List<Element> onMessageElements = XmlUtil.elements(element, "onMessage");
    if (onMessageElements == null || onMessageElements.isEmpty()) {
      parse.addProblem("pick element " + element + "should contain at least one onMessage element.");
    }
    List<Element> onAlarmElements = XmlUtil.elements(element, "onAlarm");
    if (onAlarmElements != null && !onAlarmElements.isEmpty() && pick.isCreateInstance()) {
      String message = StaticAnalysisFault.getMessage("SA00062", "  pick " + pick
        + " has a createInstance attribute with a value of yes. It must contain only onMessage events.");
      parse.addProblem(message, new StaticAnalysisException("SA00062"));
    }
    List<OnMessage> onMessages = new ArrayList<OnMessage>();
    List<OnAlarm> onAlarms = null;
    parse.pushObject(pickNode);
    try {
      for (Element onMessageElement : onMessageElements) {
        OnMessageFullDefinition onMessageFullDefinition = new OnMessageFullDefinitionImpl();

        OnMessage onMessage = parseOnMessage(onMessageElement, pickNode, pick, parse, parser, onMessageFullDefinition);
        onMessages.add(onMessage);

        onMessageFullDefinitionList.add(onMessageFullDefinition);
      }
      if (onAlarmElements != null) {
        onAlarms = new ArrayList<OnAlarm>();
        for (Element onAlarmElement : onAlarmElements) {
          OnAlarmFullDefinition onAlarmFullDefinition = new OnAlarmFullDefinitionImpl();

          OnAlarm onAlarm = parseOnAlarm(onAlarmElement, pickNode, pick, parse, parser, onAlarmFullDefinition);
          onAlarms.add(onAlarm);

          onAlarmFullDefinitionList.add(onAlarmFullDefinition);
        }
      }
    } finally {
      parse.popObject();
    }
    pick.setOnMessages(onMessages);
    pick.setOnAlarms(onAlarms);

    pickActivityFullDefinition.setOnMessageDefinitionList(onMessageFullDefinitionList);
    pickActivityFullDefinition.setOnAlarmDefinitionList(onAlarmFullDefinitionList);
    pickActivityFullDefinition.setCreateInstance(pick.isCreateInstance());

    return pickNode;
  }

  public OnAlarm parseOnAlarm(final Element element, final NodeImpl pickNode, final Pick pick, final Parse parse,
    final Parser parser, final OnAlarmFullDefinition onAlarmFullDefinition) {
    if (log.isLoggable(Level.FINE)) {
      log.fine("parsing element = " + element);
    }
    OnAlarm onAlarm = new OnAlarm();

     // Definition
    DefinitionKeeper defKeeper = parse.findObject(DefinitionKeeper.class);
    defKeeper.pushParent(onAlarmFullDefinition);

    onAlarm.setActivity(parseChildActivity(element, parse, parser));

    defKeeper.popParent();

    BpelProcess bpelProcess = parse.findObject(BpelProcess.class);
    Element deadlineElement = XmlUtil.element(element, "until");
    Element durationElement = XmlUtil.element(element, "for");
    if (durationElement != null && deadlineElement != null) {
      parse.addProblem("Error in element: " + element + " only one element 'for' or 'until' should be used.");
    }
    if (durationElement != null) {
      Expression expression = parseExpression(durationElement, bpelProcess);
      onAlarm.setDurationExpression(expression);

      ForDefinition forDefinition = new ForDefinitionImpl(expression.getLanguage(), durationElement.getTextContent());

      onAlarmFullDefinition.setForDefinition(forDefinition);
    } else if (deadlineElement != null) {
      Expression expression = parseExpression(deadlineElement, bpelProcess);
      onAlarm.setDeadlineExpression(expression);

      UntilDefinition untilDefinition = new UntilDefinitionImpl(expression.getLanguage(), deadlineElement.getTextContent());
      onAlarmFullDefinition.setUntilDefinition(untilDefinition);
    } else {
      parse.addProblem("Error in element: " + element + " one element 'for' or 'until' should be used.");
    }

    return onAlarm;
  }

  public OnMessage parseOnMessage(final Element element, final NodeImpl pickNode, final Pick pick, final Parse parse,
    final Parser parser, final OnMessageFullDefinition onMessageFullDefinition) {
    if (log.isLoggable(Level.FINE)) {
      log.fine("parsing element = " + element);
    }
    OnMessage onMessage = new OnMessage();
    String variable = XmlUtil.attribute(element, "variable");
    String partnerLinkName = XmlUtil.attribute(element, "partnerLink");
    String operation = XmlUtil.attribute(element, "operation");

    onMessage.setVariable(variable);
    onMessage.setPartnerLinkName(partnerLinkName);
    onMessage.setOperation(operation);

    // Definition
    onMessageFullDefinition.setOperation(operation);
    onMessageFullDefinition.setPartnerLink(partnerLinkName);
    onMessageFullDefinition.setVariable(variable);

    String messageExchange = XmlUtil.attribute(element, "messageExchange");
    if (messageExchange != null) {
      onMessage.setMessageExchange(messageExchange);
      onMessageFullDefinition.setMessageExchange(messageExchange);
    }

    // get partnerLinkDef and get plnkType attribute and myRole attribute
    PartnerLink partnerLink = findPartnerLinkDefinition(parse, partnerLinkName);
    QName plportType = partnerLink.getMyRolePortTypeQName();
    onMessage.setPortTypeQName(plportType);

    if (XmlUtil.attribute(element, "portType") != null) {
      if (!plportType.equals(XmlUtil.attributeQName(element, "portType"))) {
        // SA00005
        String message = StaticAnalysisFault.getMessage("SA00005", "  onMessage " + onMessage
          + " has specified a PortType that is different from the one specified by the partnerLink");
        parse.addProblem(message, new StaticAnalysisException("SA00005"));
      }
    }

    // Definition
    onMessageFullDefinition.setPortType(onMessage.getPortTypeQName());

    // fromParts
    Element fromPartsElement = XmlUtil.element(element, "fromParts");
    if (fromPartsElement != null) {
      if (onMessage.getVariable() != null) {
        // SA00055
        String message = StaticAnalysisFault.getMessage("SA00063", "  onMessage " + onMessage
          + " has specified both a Variable and a fromPart");
        parse.addProblem(message, new StaticAnalysisException("SA00063"));
      }
      List<Element> fromPartElements = XmlUtil.elements(fromPartsElement, "fromPart");
      onMessage.setAnonymousInboundMessageVariable(ReceiveBinding.getAnonymousVariable(onMessage, parse));
      List<Copy> inCopies = ReceiveBinding.getInCopies(fromPartElements, pick.getEnclosingScope(), parse,
        onMessageFullDefinition);
      if (inCopies.size() != 0) {
        onMessage.setInCopies(inCopies);
      }
    }

    // Definition
    DefinitionKeeper defKeeper = parse.findObject(DefinitionKeeper.class);
    defKeeper.pushParent(onMessageFullDefinition);

    // correlations
    Element correlationsElement = XmlUtil.element(element, "correlations");
    if (correlationsElement != null) {
      List<Element> correlationElements = XmlUtil.elements(correlationsElement, "correlation");
      onMessage.setCorrelations(ReceiveBinding.parseCorrelations(correlationElements, pick.getEnclosingScope(), parse,
        onMessageFullDefinition));
    }
    BpelProcess bpelProcess = parse.findObject(BpelProcess.class);

    if (pick.isCreateInstance()) {
      OperationKey operationKey = onMessage.getOperationKey(bpelProcess.getQName());
      bpelProcess.addStartElement(operationKey, onMessage, correlationsElement != null);
    }

    onMessage.setActivity(parseChildActivity(element, parse, parser));

    defKeeper.popParent();

    return onMessage;
  }
  protected static Expression parseExpression(final Element element, final BpelProcess bpelProcess) {
    String expressionLanguage = XmlUtil.attribute(element, "expressionLanguage");
    if (expressionLanguage == null) {
      expressionLanguage = bpelProcess.getExpressionLanguage();
    }
    Expression expression = new Expression();
    expression.setLanguage(expressionLanguage);
    expression.setNamespaces(BpelUtil.getAllNameSpaces(element));
    expression.setText(element.getTextContent());
    return expression;
  }

  @Override
  public ActivityFullDefinition createActivityFullDefinition(final ProcessDefinitionUUID processDefinitionUUID,
    final ActivityDefinitionUUID activityDefinitionUUID, final String name, final Parse parse) {
    return new PickActivityFullDefinitionImpl(processDefinitionUUID, activityDefinitionUUID, name);
  }
}
