/**
 * Copyright (C) 2007  Bull S. A. S.
 * Bull, Rue Jean Jaures, B.P.68, 78340, Les Clayes-sous-Bois
 * This library is free software; you can redistribute it and/or modify it under the terms
 * of the GNU Lesser General Public License as published by the Free Software Foundation
 * version 2.1 of the License.
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 * You should have received a copy of the GNU Lesser General Public License along with this
 * program; if not, write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth
 * Floor, Boston, MA  02110-1301, USA.
 **/
package org.ow2.orchestra.persistence.log;

import java.util.logging.Level;
import java.util.logging.Logger;

import org.ow2.orchestra.facade.data.runtime.ActivityInstanceData;
import org.ow2.orchestra.facade.data.runtime.CorrelationSetInitializedData;
import org.ow2.orchestra.facade.data.runtime.ForeachInitializedData;
import org.ow2.orchestra.facade.data.runtime.PartnerLinkUpdateData;
import org.ow2.orchestra.facade.data.runtime.ProcessInstanceData;
import org.ow2.orchestra.facade.data.runtime.TimerData;
import org.ow2.orchestra.facade.data.runtime.TransitionConditionUpdateData;
import org.ow2.orchestra.facade.data.runtime.VariableUpdateData;
import org.ow2.orchestra.facade.def.full.ProcessFullDefinition;
import org.ow2.orchestra.facade.runtime.full.ProcessFullInstance;
import org.ow2.orchestra.facade.uuid.ProcessDefinitionUUID;
import org.ow2.orchestra.services.Archivable;
import org.ow2.orchestra.services.Recorder;

/**
 *
 * @author Gael Lalire
 */
public class LoggerRecorder implements Recorder {

  private static final Logger LOG = Logger.getLogger(LoggerRecorder.class.getName());

  private final Level level;

  public LoggerRecorder() {
    this(Level.FINE);
  }

  public LoggerRecorder(final String levelName) {
    this(Level.parse(levelName));
  }

  public LoggerRecorder(final Level level) {
    this.level = level;
  }

  public void recordInstanceStarted(final ProcessInstanceData processInstanceData) {
    if (LOG.isLoggable(level)) {
      LOG.log(level, "Start [uuid:" + processInstanceData.getUUID() + "] an instance of process named '{"
        + processInstanceData.getProcessDefinitionData().getNameSpace() + "}"
        + processInstanceData.getProcessDefinitionData().getName() + "'");
    }
  }

  public void recordInstanceEnded(final ProcessInstanceData processInstanceData) {
    if (LOG.isLoggable(level)) {
      LOG.log(level, "Instance [uuid:" + processInstanceData.getUUID() + "] ended");
    }
  }

  public void recordActivityStarted(final ActivityInstanceData activityInstanceData) {
    if (LOG.isLoggable(level)) {
      LOG.log(level, "Start [uuid:" + activityInstanceData.getUUID() + "] an activity of an instance of process named {"
        + activityInstanceData.getActivityDefinitionData().getProcessDefinitionData().getNameSpace() + "}"
        + activityInstanceData.getActivityDefinitionData().getProcessDefinitionData().getName() + ", this "
        + activityInstanceData.getActivityDefinitionData().getType() + " activity name is '"
        + activityInstanceData.getActivityDefinitionData().getName() + "'");
    }
  }

  public void recordActivityEnded(final ActivityInstanceData activityInstanceData) {
    if (LOG.isLoggable(level)) {
      LOG.log(level, "Activity [uuid:" + activityInstanceData.getUUID() + "] ended");
    }
  }

  public void remove(final Archivable archivable) {
    if (archivable instanceof ProcessFullInstance) {
      remove((ProcessFullInstance) archivable);
    } else if (archivable instanceof ProcessFullDefinition) {
      remove((ProcessFullDefinition) archivable);
    }
  }

  public void remove(final ProcessFullInstance processFullInstance) {
    if (LOG.isLoggable(this.level)) {
      LOG.log(this.level, "ProcessInstance " + processFullInstance.getUUID() + " removed.");
    }
  }

  public void remove(final ProcessFullDefinition processFullDefinition) {
    if (LOG.isLoggable(this.level)) {
      LOG.log(this.level, "Process " + processFullDefinition.getUUID() + " removed.");
    }
  }

  public void recordActivityPending(final ActivityInstanceData activityInstanceData) {

    if (LOG.isLoggable(level)) {
      LOG.log(level, "Activity [uuid:" + activityInstanceData.getUUID() + "] pending");
    }
  }

  public void recordTimerStarted(final TimerData timerData) {
    if (LOG.isLoggable(level)) {
      LOG.log(level, "Timer [expectedStartedDate: " + timerData.getExpectedStartedDate() + "] started for activity "
        + timerData.getActivityInstanceUUID());
    }
  }

  public void recordForeachInitialized(final ForeachInitializedData foreachInitializedData) {
    if (LOG.isLoggable(level)) {
      LOG.log(level, "Foreach [startValue: " + foreachInitializedData.getStartValue() + "]" + "[stopValue: "
        + foreachInitializedData.getStopValue() + "]");
    }
  }

  public void recordVariableUpdated(final VariableUpdateData variableUpdateData) {
    if (LOG.isLoggable(level)) {
      LOG.log(level, "Variable [name: " + variableUpdateData.getName() + ", scopeUUID: "
        + variableUpdateData.getScopeActivityUUID() + "] is set to <<<" + variableUpdateData.getValue() + ">> by activity "
        + variableUpdateData.getActivityUUID());
    }
  }

  public void recordCorrelationSetInitialized(final CorrelationSetInitializedData correlationSetInitializedData) {
    if (LOG.isLoggable(level)) {
      LOG.log(level, "Correlation set [name: " + correlationSetInitializedData.getName() + ", scopeUUID: "
        + correlationSetInitializedData.getScopeActivityUUID() + "] is initialized by activity "
        + correlationSetInitializedData.getActivityUUID());
    }

  }

  public void recordPartnerLinkUpdated(final PartnerLinkUpdateData partnerLinkUpdateData) {
    if (LOG.isLoggable(level)) {
      LOG.log(level, "Partner Link [name: " + partnerLinkUpdateData.getName() + ", scopeUUID: "
        + partnerLinkUpdateData.getScopeActivityUUID() + "] is updated by activity "
        + partnerLinkUpdateData.getActivityUUID());
    }
  }

  public void recordTransitionConditionUpdated(final TransitionConditionUpdateData transitionConditionUpdateData) {
    if (LOG.isLoggable(level)) {
      LOG.log(level, "Transition Condition [name: " + transitionConditionUpdateData.getName() + ", status: "
        + transitionConditionUpdateData.getStatus() + ", flow activity UUID "
        + transitionConditionUpdateData.getFlowActivityUUID() + "]");
    }
  }

  public void recordProcessDeployed(final ProcessFullDefinition processFullDefinition) {
    if (LOG.isLoggable(level)) {
      LOG.log(level, "Deploy process [name: " + processFullDefinition.getName() + ", processUUID "
        + processFullDefinition.getUUID() + "]");
    }
  }

  public void recordProcessUndeployed(final ProcessDefinitionUUID processDefinitionUUID) {
    if (LOG.isLoggable(level)) {
      LOG.log(level, "Undeploy process [processUUID " + processDefinitionUUID + "]");
    }

  }
}
