/**
 * Copyright (C) 2006  Bull S. A. S.
 * Bull, Rue Jean Jaures, B.P.68, 78340, Les Clayes-sous-Bois
 * This library is free software; you can redistribute it and/or modify it under the terms
 * of the GNU Lesser General Public License as published by the Free Software Foundation
 * version 2.1 of the License.
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 * You should have received a copy of the GNU Lesser General Public License along with this
 * program; if not, write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth
 * Floor, Boston, MA  02110-1301, USA.
 **/
package org.ow2.orchestra.runtime;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import org.ow2.orchestra.definition.activity.Scope;
import org.ow2.orchestra.definition.element.Catch;
import org.ow2.orchestra.definition.element.CompensationHandler;
import org.ow2.orchestra.definition.element.CorrelationSet;
import org.ow2.orchestra.definition.element.PartnerLink;
import org.ow2.orchestra.definition.element.Variable;
import org.ow2.orchestra.facade.uuid.ActivityInstanceUUID;

/**
 * @author Charles Souillard
 */
public class ScopeRuntime {

  protected long dbid;

  private String uuid;

  /** scope definition of this scopeRuntime. */
  protected Scope scopeDefinition = null;

  /** all runtime variables of this scope. */
  protected Map<String, VariableRuntime> variables = null;

  /** all partnerLinkRuntime of this scope. */
  protected Map<String, PartnerLinkRuntime> partnerLinks = null;

  /** all correlationSetRuntime of this scope. */
  protected Map<String, CorrelationSetRuntime> correlationSets = null;

  /** all compensationHandlers of this scope. */
  protected List<CompensationHandler> compensationHandlers = new ArrayList<CompensationHandler>();

  private Catch catchHasAlreadyBeenMatched = null;

  protected ActivityInstanceUUID activityInstanceUUID;

  // repliers should be mapped to 1 single table. if that's not possible
  // with JPA mappings, change the type here to accomplish it.
  // protected Map<ReplierKey, Long> repliers;

  protected ScopeRuntime() {
  }

  /**
   * Default constructor.
   *
   * @param scopeDefinition
   *            scopeDefinition
   * @param execution
   *            execution that this scope belongs to
   */
  public ScopeRuntime(final Scope scopeDefinition, final BpelExecution execution,
      final String uuid) {
    this.activityInstanceUUID = execution.getActivityInstanceUUID();
    this.uuid = uuid;
    this.scopeDefinition = scopeDefinition;
    // scope children
    createVariables(execution);
    createPartnerLinks(execution);
    createCorrelationSets(execution);


    // TODO : create other runtime children...
  }

  /**
   * for each variable defined in scopeDefinition, create a new runtime
   * variable.
   *
   * @param execution
   *            execution that this scope belongs to
   */
  private void createVariables(final BpelExecution execution) {
    variables = new HashMap<String, VariableRuntime>();
    final Map<String, Variable> scopeVariables = scopeDefinition.getVariables();
    if (scopeVariables != null) {
      for (Variable variable : scopeVariables.values()) {
        // Object variableValue = getVariableInitialValue(variable, execution);
        variables.put(variable.getName(), new VariableRuntime(variable,
            execution, activityInstanceUUID));
      }
    }
  }

  public Collection<String> getVariableNames() {
    if (variables == null) {
      return Collections.emptySet();
    } else {
      return variables.keySet();
    }
  }

  public VariableRuntime createVariable(final BpelExecution execution,
      final Variable variable) {
    VariableRuntime variableRuntime = new VariableRuntime(variable, execution, activityInstanceUUID);
    variables.put(variable.getName(), variableRuntime);
    return variableRuntime;
  }

  public void removeVariable(final BpelExecution execution, final Variable variable) {
    variables.remove(variable.getName());
  }

  /**
   * for each partnerLink defined in scopeDefinition, create a new runtime
   * partnerLink.
   *
   * @param execution
   *            execution that this scope belongs to
   */
  private void createPartnerLinks(final BpelExecution execution) {
    partnerLinks = new HashMap<String, PartnerLinkRuntime>();
    final Map<String, PartnerLink> scopePartnerLinks = scopeDefinition
    .getPartnerLinks();
    if (scopePartnerLinks != null) {
      for (PartnerLink partnerLink : scopePartnerLinks.values()) {
        PartnerLinkRuntime partnerLinkRuntime = new PartnerLinkRuntime(
            partnerLink, execution, execution.getActivityInstanceUUID());
        partnerLinks.put(partnerLink.getName(), partnerLinkRuntime);
      }
    }
  }

  /**
   * for each correlationSet defined in scopeDefinition, create a new runtime
   * correlationSet.
   *
   * @param execution
   *            execution that this scope belongs to
   */
  private void createCorrelationSets(final BpelExecution execution) {
    correlationSets = new HashMap<String, CorrelationSetRuntime>();
    Map<String, CorrelationSet> correlationSetsMap = scopeDefinition
    .getCorrelationSets();
    if (correlationSetsMap != null) {
      for (CorrelationSet correlationSet : correlationSetsMap.values()) {
        CorrelationSetRuntime correlationSetRuntime = new CorrelationSetRuntime(
            correlationSet, execution.getActivityInstanceUUID());
        correlationSets.put(correlationSet.getName(), correlationSetRuntime);
      }
    }
  }


  /**
   * add a new compensation handlers.
   * @param ch the new compensation handler
   */
  public void addCompensationHandlersInCurrentScope(final List<CompensationHandler> chs) {
    if(chs != null) {
      for (CompensationHandler compensationHandler : chs) {
        if ((compensationHandler != null)&&(!this.compensationHandlers.contains(compensationHandler))) {
          this.compensationHandlers.add(0, compensationHandler);
        }
      }
    }
  }

  /**
   * add a new compensation handler.
   * @param ch the new compensation handler
   */
  public void addCompensationHandlerInCurrentScope(final CompensationHandler ch) {
    if(ch != null) {
      this.compensationHandlers.add(0, ch);
    }
  }

  public Scope getScopeDefinition() {
    return this.scopeDefinition;
  }

  public void setScopeDefinition(final Scope scopeDefinition) {
    this.scopeDefinition = scopeDefinition;
  }

  /**
   * Returns the partnerLinkRuntime having the given name.
   *
   * @param partnerLinkName
   *            partnerLinkName
   * @return the partnerLinkRuntime having the given name.
   */
  public PartnerLinkRuntime getPartnerLink(final String partnerLinkName) {
    return partnerLinks.get(partnerLinkName);
  }

  /**
   * return true if this scope contains a csRuntime with the given csName.
   *
   * @param csName
   *            csName
   * @return true if this scope contains a csRuntime with the given csName.
   */
  public boolean containsCorrelationSet(final String csName) {
    return correlationSets.get(csName) != null;
  }

  /**
   * Returns the CorrelationSetRuntime having the given name.
   *
   * @param csName
   *            csName
   * @return the CorrelationSetRuntime having the given name.
   */
  public CorrelationSetRuntime getCorrelationSet(final String csName) {
    return correlationSets.get(csName);
  }

  /**
   * Returns the VariableRuntime having the given name.
   *
   * @param variableName
   *            variableName
   * @param execution
   *            execution
   * @return the VariableRuntime having the given name.
   */
  public VariableRuntime getVariable(final String variableName,
      final BpelExecution execution) {
    return variables.get(variableName);
  }

  /**
   * return true if this scope contains a variableRuntime with the given
   * variableName.
   *
   * @param variableName
   *            variableName
   * @return true if this scope contains a variableRuntime with the given
   *         variableName.
   */
  public boolean containsVariable(final String variableName) {
    return this.variables.keySet().contains(variableName);
  }

  public String getUuid() {
    return uuid;
  }

  @Override
  public int hashCode() {
    return uuid.hashCode();
  }

  @Override
  public boolean equals(final Object obj) {
    if (obj == null) {
      return false;
    }
    if (!(obj instanceof ScopeRuntime)) {
      return false;
    }
    ScopeRuntime other = (ScopeRuntime) obj;
    return other.getUuid().equals(getUuid());
  }

  public List<CompensationHandler> getCompensationHandlers() {
    return this.compensationHandlers;
  }

  public Catch isCatchHasAlreadyBeenMatched() {
    return catchHasAlreadyBeenMatched;
  }

  public void setCatchHasAlreadyBeenMatched(final Catch catchHasAlreadyBeenMatched) {
    this.catchHasAlreadyBeenMatched = catchHasAlreadyBeenMatched;
  }

}
