/**
 * Copyright (C) 2006  Bull S. A. S.
 * Bull, Rue Jean Jaures, B.P.68, 78340, Les Clayes-sous-Bois
 * This library is free software; you can redistribute it and/or modify it under the terms
 * of the GNU Lesser General Public License as published by the Free Software Foundation
 * version 2.1 of the License.
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 * You should have received a copy of the GNU Lesser General Public License along with this
 * program; if not, write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth
 * Floor, Boston, MA  02110-1301, USA.
 **/
package org.ow2.orchestra.runtime;

import java.util.logging.Level;
import java.util.logging.Logger;

import javax.wsdl.Part;

import org.ow2.orchestra.definition.element.Copy;
import org.ow2.orchestra.definition.element.Variable;
import org.ow2.orchestra.facade.data.runtime.VariableUpdateData;
import org.ow2.orchestra.facade.exception.OrchestraRuntimeException;
import org.ow2.orchestra.facade.uuid.ActivityInstanceUUID;
import org.ow2.orchestra.util.BpelUtil;
import org.ow2.orchestra.util.BpelXmlUtil;
import org.ow2.orchestra.util.EnvTool;
import org.ow2.orchestra.var.Message;
import org.w3c.dom.Element;

/**
 * @author Charles Souillard
 */
public class VariableRuntime {

  protected long dbid;

  /** log. */
  private static Logger log = Logger.getLogger(VariableRuntime.class.getName());

  /** variable name. */
  protected String name;

  /** variable value. */
  protected Message messageValue;

  protected Element xmlValue;

  /** variable definition. */
  protected Variable variableDefinition;

  protected ActivityInstanceUUID scopeActivityInstanceUUID;

  protected VariableRuntime() {
  }

  /**
   * Default constructor.
   *
   * @param variableDefinition
   *            variableDefinition
   * @param execution
   *            execution
   */
  public VariableRuntime(final Variable variableDefinition, final BpelExecution execution,
    final ActivityInstanceUUID scopeActivityInstanceUUID) {
    this.scopeActivityInstanceUUID = scopeActivityInstanceUUID;
    this.name = variableDefinition.getName();
    this.variableDefinition = variableDefinition;
    if (this.variableDefinition.getMessageType() != null) {
      javax.wsdl.Message message = execution.getMessage(this.variableDefinition
          .getMessageType());
      messageValue = new Message(message);
    }
  }

  public void initializeXmlType(final BpelExecution execution) {
    Copy initCopy = variableDefinition.getVirtualCopy();
    if (initCopy != null) {
      initCopy.execute(execution);
    } else {
      if (variableDefinition.getXmlType() != null) {
        xmlValue = BpelXmlUtil.createDefaultElement();
      } else if (variableDefinition.getElement() != null) {
        xmlValue = BpelUtil.getDocumentWithOneElement(variableDefinition
            .getElement());
      } else {
        if (log.isLoggable(Level.FINE)) {
          log
              .fine("initializeXmlType : variable is not a type and is not an element... : this case is forbiden");
        }
      }
    }
    EnvTool.getRecorder()
      .recordVariableUpdated(
        new VariableUpdateData(execution.getActivityInstanceUUID(), execution.getActivityInstanceUUID(), name,
          xmlValue));
  }

  /**
   * Initialize the variable.
   *
   * @param part :
   *            part of the message to initialize, if the variable is a message.
   *            Else null
   * @param execution
   *            execution
   */
  public void initializePart(final BpelExecution execution, final String part) {
    if (variableDefinition.getMessageType() != null) {
      if (part != null) {
        // we have to initialize only one part
        javax.wsdl.Message message = execution.getMessage(variableDefinition
            .getMessageType());
        Part partRef = message.getPart(part);
        // value = new Message(message);
        if (partRef.getTypeName() != null) {
          // it is a simple type or a complex type part: we initialize it
          // with a single anonymous Element
          messageValue.setPart(part, BpelXmlUtil.createDefaultElement());
        } else {
          // it is an element part.
          messageValue.setPart(part, BpelUtil.getDocumentWithOneElement(partRef
              .getElementName()));
        }
      }
    } else {
      if (log.isLoggable(Level.FINE)) {
        log
            .fine("initializePart : variable is not a message... : this case is forbiden");
      }
    }
    EnvTool.getRecorder().recordVariableUpdated(
      new VariableUpdateData(execution.getActivityInstanceUUID(), execution.getActivityInstanceUUID(), name, messageValue));
  }

  /**
   * This method must be called only if the variable is not a message.
   *
   * @return true is the variable is already initialized
   */
  public boolean isXmlInitialized() {
    if (variableDefinition.getMessageType() != null) {
      throw new OrchestraRuntimeException("Variable " + name
          + " is a message, we cannot call " + this.getClass().getName()
          + ".isXmlInitialized method."
          + "please call isPartInitialized method");
    }
    return xmlValue != null;
  }

  /**
   * This method must be called only if the variable is a message.
   *
   * @param partName
   *            name of the part we want to know if it is initialized or not
   * @return true is the given part is already initialized
   */
  public boolean isPartInitialized(final String partName) {
    if (variableDefinition.getMessageType() == null) {
      throw new OrchestraRuntimeException("Variable " + name
          + " is not a message, we cannot call " + this.getClass().getName()
          + ".isPartInitialized method."
          + "please call isXmlInitialized method");
    }
    if (messageValue == null) {
      return false;
    } else {
      return messageValue.isPartInitialized(partName);
    }
  }

  public Variable getVariableDefinition() {
    return this.variableDefinition;
  }

  public void setVariableDefinition(final Variable variableDefinition) {
    this.variableDefinition = variableDefinition;
  }

  public String getName() {
    return this.name;
  }

  public void setName(final String name) {
    this.name = name;
  }

  public Object getValue() {
    if (variableDefinition.getMessageType() != null) {
      return messageValue;
    }
    return xmlValue;
  }

  public void setValue(final Message messageValue, final ActivityInstanceUUID activityInstanceUUID) {
    if (variableDefinition.getMessageType() == null) {
      throw new OrchestraRuntimeException("trying to set a messageValue and "
          + "the variableDefinition is not a messageType : " + this.toString());
    } else {
      this.messageValue = messageValue;
    }
    EnvTool.getRecorder().recordVariableUpdated(
      new VariableUpdateData(scopeActivityInstanceUUID, activityInstanceUUID, name, messageValue));
  }

  public void setValue(final Object value, final boolean keepSrcElementName, final ActivityInstanceUUID activityInstanceUUID) {
    if (variableDefinition.getMessageType() != null) {
      throw new OrchestraRuntimeException("trying to set a xmlValue and the variableDefinition is a messageType : "
        + this.toString());
    }
    if (xmlValue == null) {
      if (variableDefinition.getXmlType() != null) {
        xmlValue = BpelXmlUtil.createDefaultElement();
      } else if (variableDefinition.getElement() != null) {
        xmlValue = BpelUtil.getDocumentWithOneElement(variableDefinition.getElement());
      } else {
        throw new OrchestraRuntimeException("trying to set a xmlValue and the variableDefinition type is unknown.");
      }
    }
    BpelXmlUtil.setObjectValue(xmlValue, value, keepSrcElementName);
    EnvTool.getRecorder().recordVariableUpdated(
      new VariableUpdateData(scopeActivityInstanceUUID, activityInstanceUUID, name, xmlValue));
  }

  public long getDbid() {
    return dbid;
  }

//  public String toString() {
//    return VariableRuntime.class.getName()
//        + "[dbid: " + dbid
//        + ", name: " + name
//        + ", messageValue:" + messageValue
//        + ", xmlValue: " + xmlValue
//        + ", variableDefinition: " + variableDefinition
//        + "]";
//  }
}
