/**
 * Copyright (C) 2006  Bull S. A. S.
 * Bull, Rue Jean Jaures, B.P.68, 78340, Les Clayes-sous-Bois
 * This library is free software; you can redistribute it and/or modify it under the terms
 * of the GNU Lesser General Public License as published by the Free Software Foundation
 * version 2.1 of the License.
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 * You should have received a copy of the GNU Lesser General Public License along with this
 * program; if not, write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth
 * Floor, Boston, MA  02110-1301, USA.
 **/
package org.ow2.orchestra.services.impl;

import java.util.HashSet;
import java.util.List;
import java.util.Set;

import org.jbpm.pvm.activity.Activity;
import org.jbpm.pvm.env.Environment;
import org.jbpm.pvm.internal.model.NodeImpl;
import org.jbpm.pvm.model.Node;
import org.ow2.orchestra.definition.BpelProcess;
import org.ow2.orchestra.definition.element.InboundMessageElement;
import org.ow2.orchestra.services.OperationKey;
import org.ow2.orchestra.services.Receiver;
import org.ow2.orchestra.services.ReceivingElement;
import org.ow2.orchestra.services.itf.Publisher;
import org.ow2.orchestra.services.itf.Repository;

public class PublisherImpl implements Publisher {

  private Repository repository = null;

  private synchronized Repository getRepository() {
    if (repository == null) {
      repository = Environment.getCurrent().get(Repository.class);
    }
    return repository;
  }

  public Receiver getReceiver(final OperationKey operationKey) {
    return getRepository().getReceiver(operationKey);
  }

  public Receiver removeReceiver(final OperationKey operationKey) {
    return getRepository().removeReceiver(operationKey);
  }

  public void storeReceiver(final OperationKey operationKey, final Receiver receiver) {
    getRepository().storeReceiver(operationKey, receiver);
  }

  public void publishServices(final BpelProcess bpelProcess, final Environment environment) {
    // set of unique operation keys defined in the process
    Set<OperationKey> operationKeys = getServiceOperationKeys(bpelProcess, bpelProcess.getInitial(),
        environment);
    // for each operation key, deploy a receiver
    for (OperationKey operationKey : operationKeys) {
      Receiver receiver = createReceiverInstance(operationKey);
      storeReceiver(operationKey, receiver);
    }
  }

  protected Set<OperationKey> getServiceOperationKeys(final BpelProcess bpelProcess, final NodeImpl node,
      final Environment environment) {

    Set<OperationKey> operationKeys = new HashSet<OperationKey>();
    // get list of operation keys for the current node
    Activity behaviour = node.getBehaviour();
    if (behaviour instanceof InboundMessageElement) {
      InboundMessageElement inboundMessageElement = (InboundMessageElement) behaviour;
      for (ReceivingElement receivingElement : inboundMessageElement.getReceivingElements()) {
        OperationKey operationKey = receivingElement.getOperationKey(bpelProcess
            .getQName());
        operationKeys.add(operationKey);
      }
    }
    // Add list of operation keys defined in children nodes
    List<Node> children = node.getNodes();
    if (children != null) {
      for (Node child : children) {
        operationKeys.addAll(getServiceOperationKeys(bpelProcess, (NodeImpl) child, environment));
      }
    }

    return operationKeys;
  }

  public void unpublishServices(final BpelProcess bpelProcess, final Environment environment) {
    unpublishServices(bpelProcess, bpelProcess.getInitial(),
        environment);
  }

  public void unpublishServices(final BpelProcess bpelProcess, final NodeImpl node,
      final Environment environment) {
    Activity behaviour = node.getBehaviour();
    if (behaviour instanceof InboundMessageElement) {
      InboundMessageElement inboundMessageElement = (InboundMessageElement) behaviour;
      for (ReceivingElement receivingElement : inboundMessageElement.getReceivingElements()) {
        OperationKey operationKey = receivingElement.getOperationKey(bpelProcess
            .getQName());
        removeReceiver(operationKey);
      }
    }
    List<Node> children = node.getNodes();
    if (children != null) {
      for (Node child : children) {
        unpublishServices(bpelProcess, (NodeImpl) child, environment);
      }
    }
  }

  public void republishServices(final BpelProcess bpelProcess, final Environment environment) {
  }

  protected Receiver createReceiverInstance(final OperationKey operationKey) {
    return new Receiver(operationKey);
  }

}
