/**
 * Copyright (C) 2007  Bull S. A. S.
 * Bull, Rue Jean Jaures, B.P.68, 78340, Les Clayes-sous-Bois
 * This library is free software; you can redistribute it and/or modify it under the terms
 * of the GNU Lesser General Public License as published by the Free Software Foundation
 * version 2.1 of the License.
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 * You should have received a copy of the GNU Lesser General Public License along with this
 * program; if not, write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth
 * Floor, Boston, MA  02110-1301, USA.
 **/
package org.ow2.orchestra.services.impl;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.Comparator;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.logging.Level;
import java.util.logging.Logger;

import javax.xml.namespace.QName;

import org.ow2.orchestra.facade.Record;
import org.ow2.orchestra.facade.def.ProcessState;
import org.ow2.orchestra.facade.def.full.ActivityFullDefinition;
import org.ow2.orchestra.facade.def.full.ProcessFullDefinition;
import org.ow2.orchestra.facade.runtime.ActivityState;
import org.ow2.orchestra.facade.runtime.ProcessInstanceState;
import org.ow2.orchestra.facade.runtime.full.ActivityFullInstance;
import org.ow2.orchestra.facade.runtime.full.ProcessFullInstance;
import org.ow2.orchestra.facade.uuid.AbstractUUID;
import org.ow2.orchestra.facade.uuid.ActivityDefinitionUUID;
import org.ow2.orchestra.facade.uuid.ProcessDefinitionUUID;
import org.ow2.orchestra.facade.uuid.ProcessInstanceUUID;
import org.ow2.orchestra.services.Querier;

/**
 * @author Guillaume Porcher
 *
 *         Chainer for Queriers : - for methods that returns only one object,
 *         search for the first matching object - for methods that returns a
 *         collection: search for all objects in all return queriers.
 */
public class QuerierChainer implements Querier {

  private static Logger log = Logger.getLogger(QuerierChainer.class.getName());

  private final List<Querier> queriers;

  private <E extends Record> void putRecordCollection(final Map<AbstractUUID, E> map, final Collection<E> recordCollection) {
    if (recordCollection != null) {
      for (E record : recordCollection) {
        AbstractUUID uuid = record.getUUID();
        if (map.put(uuid, record) != null) {
          if (log.isLoggable(Level.WARNING)) {
            log.warning("Duplicated entry (same uuid:" + uuid + ") found keep only the newest");
          }
        }
      }
    }
  }

  private static class DeploymentDateComparator implements Comparator<ProcessFullDefinition> {

    public int compare(final ProcessFullDefinition o1, final ProcessFullDefinition o2) {
      long o1date = o1.getDeployedDate().getTime();
      long o2date = o2.getDeployedDate().getTime();
      if (o1date == o2date) {
        return 0;
      } else if (o1date < o2date) {
        return -1;
      } else {
        return 1;
      }
    }

  }

  public QuerierChainer(final List<Querier> queriers) {
    this.queriers = queriers;
  }

  public Set<ActivityFullInstance> findActivityInstances(final ProcessInstanceUUID processUUID) {
    Map<AbstractUUID, ActivityFullInstance> activityFullInstanceByUUIDMap = new HashMap<AbstractUUID, ActivityFullInstance>();
    final Set<ActivityFullInstance> activityInsts = new HashSet<ActivityFullInstance>();
    for (final Querier querier : this.queriers) {
      putRecordCollection(activityFullInstanceByUUIDMap, querier.findActivityInstances(processUUID));
    }
    activityInsts.addAll(activityFullInstanceByUUIDMap.values());
    return activityInsts;
  }

  public Set<ActivityFullInstance> findActivityInstances(final ProcessInstanceUUID processUUID,
      final ActivityState activityState) {
    Map<AbstractUUID, ActivityFullInstance> activityFullInstanceByUUIDMap = new HashMap<AbstractUUID, ActivityFullInstance>();
    final Set<ActivityFullInstance> activityInsts = new HashSet<ActivityFullInstance>();
    for (final Querier querier : this.queriers) {
      putRecordCollection(activityFullInstanceByUUIDMap, querier.findActivityInstances(processUUID, activityState));
    }
    activityInsts.addAll(activityFullInstanceByUUIDMap.values());
    return activityInsts;
  }

  public Set<ProcessFullInstance> findProcessInstances() {
    Map<AbstractUUID, ProcessFullInstance> processFullInstanceByUUIDMap = new HashMap<AbstractUUID, ProcessFullInstance>();
    final Set<ProcessFullInstance> processInsts = new HashSet<ProcessFullInstance>();
    for (final Querier querier : this.queriers) {
      putRecordCollection(processFullInstanceByUUIDMap, querier.findProcessInstances());
    }
    processInsts.addAll(processFullInstanceByUUIDMap.values());
    return processInsts;
  }

  public Set<ProcessFullInstance> findProcessInstances(final ProcessDefinitionUUID processDefinitionUUID) {
    Map<AbstractUUID, ProcessFullInstance> processFullInstanceByUUIDMap = new HashMap<AbstractUUID, ProcessFullInstance>();
    final Set<ProcessFullInstance> processInsts = new HashSet<ProcessFullInstance>();
    for (final Querier querier : this.queriers) {
      putRecordCollection(processFullInstanceByUUIDMap, querier.findProcessInstances(processDefinitionUUID));
    }
    processInsts.addAll(processFullInstanceByUUIDMap.values());
    return processInsts;
  }

  public Set<ProcessFullInstance> findProcessInstances(final ProcessDefinitionUUID processDefinitionUUID,
      final ProcessInstanceState processInstanceState) {
    Map<AbstractUUID, ProcessFullInstance> processFullInstanceByUUIDMap = new HashMap<AbstractUUID, ProcessFullInstance>();
    final Set<ProcessFullInstance> processInsts = new HashSet<ProcessFullInstance>();
    for (final Querier querier : this.queriers) {
      putRecordCollection(processFullInstanceByUUIDMap,
        querier.findProcessInstances(processDefinitionUUID, processInstanceState));
    }
    processInsts.addAll(processFullInstanceByUUIDMap.values());
    return processInsts;
  }

  public Set<ProcessFullInstance> findProcessInstances(final ProcessInstanceState processInstanceState) {
    Map<AbstractUUID, ProcessFullInstance> processFullInstanceByUUIDMap = new HashMap<AbstractUUID, ProcessFullInstance>();
    final Set<ProcessFullInstance> processInsts = new HashSet<ProcessFullInstance>();
    for (final Querier querier : this.queriers) {
      putRecordCollection(processFullInstanceByUUIDMap, querier.findProcessInstances(processInstanceState));
    }
    processInsts.addAll(processFullInstanceByUUIDMap.values());
    return processInsts;
  }

  public ProcessFullInstance getProcessInstance(final ProcessInstanceUUID processInstanceUUID) {
    for (final Querier querier : this.queriers) {
      ProcessFullInstance processFullInstance =
        querier.getProcessInstance(processInstanceUUID);
      if (processFullInstance != null) {
        return processFullInstance;
      }
    }
    return null;
  }

  public ActivityFullDefinition getActivityDefinition(final ActivityDefinitionUUID activtyDefinitionUUID) {
    for (final Querier querier : this.queriers) {
      ActivityFullDefinition activityFullInstance = querier.getActivityDefinition(activtyDefinitionUUID);
      if (activityFullInstance != null) {
        return activityFullInstance;
      }
    }
    return null;
  }

  public List<ProcessFullDefinition> findProcessDefinitions(final QName name, final ProcessState state) {
    Map<AbstractUUID, ProcessFullDefinition> processFullInstanceByUUIDMap = new HashMap<AbstractUUID, ProcessFullDefinition>();
    // uniqueness per uuid
    for (final Querier querier : this.queriers) {
      putRecordCollection(processFullInstanceByUUIDMap, querier.findProcessDefinitions(name, state));
    }
    List<ProcessFullDefinition> processFullDefinitionList = new ArrayList<ProcessFullDefinition>();
    processFullDefinitionList.addAll(processFullInstanceByUUIDMap.values());
    // sort the list
    Collections.sort(processFullDefinitionList, new DeploymentDateComparator());

    return processFullDefinitionList;
  }

  public List<ProcessFullDefinition> findProcessDefinitions(final QName name) {
    Map<AbstractUUID, ProcessFullDefinition> processFullInstanceByUUIDMap = new HashMap<AbstractUUID, ProcessFullDefinition>();
    // uniqueness per uuid
    for (final Querier querier : this.queriers) {
      putRecordCollection(processFullInstanceByUUIDMap, querier.findProcessDefinitions(name));
    }
    List<ProcessFullDefinition> processFullDefinitionList = new ArrayList<ProcessFullDefinition>();
    processFullDefinitionList.addAll(processFullInstanceByUUIDMap.values());
    // sort the list
    Collections.sort(processFullDefinitionList, new DeploymentDateComparator());

    return processFullDefinitionList;
  }

  public List<ProcessFullDefinition> findProcessDefinitions(final ProcessState state) {
    Map<AbstractUUID, ProcessFullDefinition> processFullInstanceByUUIDMap = new HashMap<AbstractUUID, ProcessFullDefinition>();
    // uniqueness per uuid
    for (final Querier querier : this.queriers) {
      putRecordCollection(processFullInstanceByUUIDMap, querier.findProcessDefinitions(state));
    }
    List<ProcessFullDefinition> processFullDefinitionList = new ArrayList<ProcessFullDefinition>();
    processFullDefinitionList.addAll(processFullInstanceByUUIDMap.values());
    // sort the list
    Collections.sort(processFullDefinitionList, new DeploymentDateComparator());

    return processFullDefinitionList;
  }

  public List<ProcessFullDefinition> findProcessDefinitions() {
    Map<AbstractUUID, ProcessFullDefinition> processFullInstanceByUUIDMap = new HashMap<AbstractUUID, ProcessFullDefinition>();
    // uniqueness per uuid
    for (final Querier querier : this.queriers) {
      putRecordCollection(processFullInstanceByUUIDMap, querier.findProcessDefinitions());
    }
    List<ProcessFullDefinition> processFullDefinitionList = new ArrayList<ProcessFullDefinition>();
    processFullDefinitionList.addAll(processFullInstanceByUUIDMap.values());
    // sort the list
    Collections.sort(processFullDefinitionList, new DeploymentDateComparator());

    return processFullDefinitionList;
  }

  public ProcessFullDefinition getProcessDefinition(final ProcessDefinitionUUID processDefinitionUUID) {
    for (final Querier querier : this.queriers) {
      ProcessFullDefinition processFullInstance = querier.getProcessDefinition(processDefinitionUUID);
      if (processFullInstance != null) {
        return processFullInstance;
      }
    }
    return null;
  }

}
