/**
 * Copyright (C) 2007  Bull S. A. S.
 * Bull, Rue Jean Jaures, B.P.68, 78340, Les Clayes-sous-Bois
 * This library is free software; you can redistribute it and/or modify it under the terms
 * of the GNU Lesser General Public License as published by the Free Software Foundation
 * version 2.1 of the License.
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 * You should have received a copy of the GNU Lesser General Public License along with this
 * program; if not, write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth
 * Floor, Boston, MA  02110-1301, USA.
 **/
package org.ow2.orchestra.util;

import java.io.File;
import java.io.IOException;
import java.util.ArrayList;
import java.util.Date;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import org.jbpm.pvm.PvmException;
import org.jbpm.pvm.env.Environment;
import org.jbpm.pvm.internal.type.Converter;
import org.jbpm.pvm.internal.type.Type;
import org.jbpm.pvm.internal.type.Variable;
import org.jbpm.pvm.internal.type.VariableTypeResolver;
import org.jbpm.pvm.internal.type.variable.NullVariable;
import org.jbpm.pvm.internal.type.variable.UnpersistableVariable;

/**
 * @author Pierre Vigneras
 * @date Mar 6, 2008
 */
public abstract class Tool {

  public static final String DEPLOYMENT_IDPREFIX = "deployment$";

  private Tool() { }

  private static Object deploymentMutex = new Object();

  public static Object getDeploymentMutex() {
    return deploymentMutex;
  }


  public static List<String> copy(final List<String> src) {
    if (src == null) {
      return new ArrayList<String>();
    }
    return new ArrayList<String>(src);
  }

  public static Map<String, String> copy(final Map<String, String> src) {
    if (src == null) {
      return new HashMap<String, String>();
    }
    return new HashMap<String, String>(src);
  }

  public static Date copy(final Date src) {
    if (src == null) {
      return null;
    }
    return (Date) src.clone();
  }

  public static Map<String, Object> getVariableValues(final Map<String, Variable> variableMap) {
    Map<String, Object> res = new HashMap<String, Object>();
    if (variableMap != null) {
      for (Variable var : variableMap.values()) {
        res.put(var.getKey(), var.getValue());
      }
    }
    return res;
  }

  public static Map<String, Variable> getCopy(final Map<String, Variable> variableMap) {
    return createVariableMap(getVariableValues(variableMap));
  }

  public static Map<String, Variable> createVariableMap(final Map<String, Object> variables) {
    if (variables == null || variables.isEmpty()) {
      return null;
    }
    Map<String, Variable> variableMap = new HashMap<String, Variable>();
    for (Map.Entry<String, Object> e : variables.entrySet()) {
      Object value = e.getValue();
      variableMap.put(e.getKey(), createVariable(e.getKey(), value));
    }
    return variableMap;
  }

  /**
   * Copied from the pvm (in VariableMap)
   * TODO: factorize code
   */
  public static Variable createVariable(final String key, final Object value) {
    Type type = null;
    Environment environment = Environment.getCurrent();
    if (environment != null) {
      VariableTypeResolver variableTypeResolver = environment.get(VariableTypeResolver.class);
      if (variableTypeResolver != null) {
        type = variableTypeResolver.findTypeByMatch(key, value);
      }
    }
    Variable variable = null;
    if (type != null) {
      Class< ? > variableClass = type.getVariableClass();
      try {
        variable = (Variable) variableClass.newInstance();
      } catch (Exception e) {
        throw new PvmException("couldn't instantiate variable instance class '" + variableClass.getName() + "'");
      }
      Converter converter = type.getConverter();
      variable.setConverter(converter);

    } else {
      if (value == null) {
        variable = new NullVariable();
      } else {
        variable = new UnpersistableVariable();
      }
    }

    variable.setKey(key);
    variable.setValue(getVariableValueCopy(value));
    return variable;
  }

  public static Object getVariableValueCopy(final Object value) {
    if (value instanceof String) {
      return value;
    } else {
      // TODO: throw exception ?
      return value;
    }
  }


  /**
   * Perform java.io.File.createTempFile with retries when it fail (limit of 10 retries)
   * (Use to by-pass bug #6325169 on SUN JDK 1.5 on windows)
   * Sames parameter as {@link java.io.File.createTempFile} method
   * @param prefix Prefix of the file
   * @param suffix Suffix of the file
   * @param directory Target directory
   * @return An abstract pathname denoting a newly-created empty file
   * @throws IOException If a file could not be created
   */
  public static File createTempFile(final String prefix, final String suffix, final File directory) throws IOException {
    // By-pass for the bug #6325169 on SUN JDK 1.5 on windows
    // The createTempFile could fail while creating a file with the same name of an existing directory
    // So if the file creation fail, it retry (with a limit of 10 retry)
    // Rethrow the IOException if all retries failed
    File tmpDir = null;
    final int retryNumber = 10;
    int j = 0;
    boolean succeded = false;
    do {
      try {
        tmpDir = File.createTempFile(prefix, suffix, directory);
        succeded = true;
      } catch (IOException e) {
        if (j == retryNumber) {
          throw e;
        }
        try {
          Thread.sleep(100);
        } catch (InterruptedException e1) { }
        j++;
      }
    } while(!succeded);
    return tmpDir;
  }
}
